import csv
import logging
import re
from enum import Enum
from typing import List, Dict

from brain_brew.representation.configuration.global_config import GlobalConfig
from brain_brew.representation.generic.yaml_file import YamlFile, ConfigKey
from brain_brew.utils import list_of_str_to_lowercase, generate_anki_guid
from brain_brew.representation.generic.generic_file import GenericFile


class CsvKeys(Enum):
    GUID = "guid"
    TAGS = "tags"


class CsvFile(GenericFile):
    file_location: str = ""
    _data: List[dict] = []

    column_headers: list = []

    def __init__(self, file, read_now=True, data_override=None):
        super(CsvFile, self).__init__(file, read_now=read_now, data_override=data_override)

    def read_file(self):
        self._data = []

        with open(self.file_location, mode='r') as csv_file:
            csv_reader = csv.DictReader(csv_file)

            self.column_headers = list_of_str_to_lowercase(csv_reader.fieldnames)

            for row in csv_reader:
                self._data.append({key.lower(): row[key] for key in row})

        self.data_state = GenericFile.DataState.READ_IN_DATA

    def write_file(self):
        with open(self.file_location, mode='w') as csv_file:
            csv_writer = csv.DictWriter(csv_file, fieldnames=self.column_headers)

            csv_writer.writeheader()

            for row in self._data:
                csv_writer.writerow(row)

        self.file_exists = True

    def set_data(self, data_override):
        super().set_data(data_override)
        self.column_headers = list(data_override[0].keys()) if data_override else []

    def set_data_from_superset(self, superset: List[dict], column_header_override=None):
        if column_header_override:
            self.column_headers = column_header_override

        data_to_set: List[dict] = []
        for row in superset:
            if not all(column in row for column in self.column_headers):
                continue
            new_row = {}
            for column in self.column_headers:
                new_row[column] = row[column]
            data_to_set.append(new_row)

        super().set_data(data_to_set)

    def get_data(self, deep_copy=False) -> List[dict]:
        return super().get_data(deep_copy=deep_copy)

    @staticmethod
    def to_filename_csv(filename: str) -> str:
        converted = re.sub(r'\s+', '-', filename).strip()
        return converted + ".csv" if not converted.endswith(".csv") else converted

    @classmethod
    def formatted_file_location(cls, location):
        return cls.to_filename_csv(location)

    def sort_data(self, sort_by_keys, reverse_sort):
        self._data = self._sort_data(self._data, sort_by_keys, reverse_sort)



from enum import Enum
from typing import List, Dict

from brain_brew.build_tasks.build_task_generic import BuildTaskGeneric
from brain_brew.constants.build_config_keys import BuildTaskEnum, BuildConfigKeys
from brain_brew.constants.deckpart_keys import DeckPartNoteKeys
from brain_brew.interfaces.verifiable import Verifiable
from brain_brew.representation.configuration.csv_file_mapping import CsvFileMapping
from brain_brew.representation.configuration.note_model_mapping import NoteModelMapping
from brain_brew.utils import single_item_to_list
from brain_brew.representation.generic.yaml_file import ConfigKey, YamlFile
from brain_brew.representation.generic.csv_file import CsvFile, CsvKeys
from brain_brew.representation.json.deck_part_notemodel import DeckPartNoteModel
from brain_brew.representation.json.deck_part_notes import DeckPartNotes


class SourceCsvKeys(Enum):
    NOTES = "notes"
    NOTE_MODEL_MAPPINGS = "note_model_mappings"
    CSV_MAPPINGS = "csv_file_mappings"


class SourceCsv(YamlFile, BuildTaskGeneric, Verifiable):
    @staticmethod
    def get_build_keys():
        return [
            BuildTaskEnum("deck_parts_to_csv_collection", SourceCsv, "deck_parts_to_source", "source_to_deck_parts"),
            BuildTaskEnum("csv_collection_to_deck_parts", SourceCsv, "source_to_deck_parts", "deck_parts_to_source"),
        ]

    config_entry = {}
    expected_keys = {
        SourceCsvKeys.NOTES.value: ConfigKey(True, str, None),
        SourceCsvKeys.NOTE_MODEL_MAPPINGS.value: ConfigKey(True, list, None),
        SourceCsvKeys.CSV_MAPPINGS.value: ConfigKey(True, list, None),
    }
    subconfig_filter = None

    notes: DeckPartNotes
    note_model_mappings_dict: Dict[str, NoteModelMapping]
    csv_file_mappings: List[CsvFileMapping]

    def __init__(self, config_data: dict, read_now=True):
        self.setup_config_with_subconfig_replacement(config_data)
        self.verify_config_entry()

        self.notes = DeckPartNotes.create(self.get_config(BuildConfigKeys.NOTES), read_now=read_now)

        nm_mapping = [NoteModelMapping(config, read_now=read_now)
                                                         for config in
                                                         self.get_config(SourceCsvKeys.NOTE_MODEL_MAPPINGS)
                                                         ]
        self.note_model_mappings_dict = {mapping.note_model.name: mapping
                                         for mapping in nm_mapping}

        self.csv_file_mappings = [CsvFileMapping(config, read_now=read_now)
                                  for config in self.get_config(SourceCsvKeys.CSV_MAPPINGS)]

    @classmethod
    def from_yaml(cls, yaml_file_name, read_now=True):
        config_data = YamlFile.read_file(yaml_file_name)

        return SourceCsv(config_data, read_now=read_now)

    def verify_contents(self):
        errors = []

        for nm in self.note_model_mappings_dict.values():
            try:
                nm.verify_contents()
            except KeyError as e:
                errors.append(e)

        for cfm in self.csv_file_mappings:
            # Check all necessary key values are present
            try:
                cfm.verify_contents()
            except KeyError as e:
                errors.append(e)

            # Check all references notemodels have a mapping
            for csv_map in self.csv_file_mappings:
                for nm in csv_map.get_used_note_model_names():
                    if nm not in self.note_model_mappings_dict.keys():
                        errors.append(f"Missing Note Model Map for {nm}")

        # Check each of the Csvs (or their derivatives) contain all the necessary columns for their stated note model
        for cfm in self.csv_file_mappings:
            note_model_names = cfm.get_used_note_model_names()
            available_columns = cfm.get_available_columns()

            referenced_note_models_maps = [value for key, value in self.note_model_mappings_dict.items() if
                                           key in note_model_names]
            for nm_map in referenced_note_models_maps:
                missing_columns = [col for col in nm_map.note_model.fields_lowercase if
                                   col not in nm_map.csv_headers_map_to_note_fields(available_columns)]
                if missing_columns:
                    errors.append(KeyError(f"Csvs are missing columns from {nm_map.note_model.name}", missing_columns))

        if errors:
            raise Exception(errors)

    def notes_to_deck_parts(self):
        csv_data_by_guid: Dict[str, dict] = {}
        for csv_map in self.csv_file_mappings:
            csv_map.compile_data()
            csv_data_by_guid = {**csv_data_by_guid, **csv_map.compiled_data}
        csv_rows: List[dict] = list(csv_data_by_guid.values())

        notes_json = []
        top_level_note_structure = {
            DeckPartNoteKeys.FIELDS.value: List[str],
            DeckPartNoteKeys.GUID.value: "",
            DeckPartNoteKeys.TAGS.value: List[str],
            DeckPartNoteKeys.NOTE_MODEL.value: ""
        }

        # Get Guid, Tags, NoteTypeName, Fields
        for row in csv_rows:
            note = top_level_note_structure.copy()

            row_nm: NoteModelMapping = self.note_model_mappings_dict[row[DeckPartNoteKeys.NOTE_MODEL.value]]

            filtered_fields = row_nm.csv_row_map_to_note_fields(row)

            note[DeckPartNoteKeys.NOTE_MODEL.value] = row_nm.note_model.name
            note[DeckPartNoteKeys.GUID.value] = filtered_fields.pop(DeckPartNoteKeys.GUID.value)
            note[DeckPartNoteKeys.TAGS.value] = self.split_tags(filtered_fields.pop(DeckPartNoteKeys.TAGS.value))

            note[DeckPartNoteKeys.FIELDS.value] = row_nm.field_values_in_note_model_order(filtered_fields)

            notes_json.append(note)

        return notes_json

    def notes_to_source(self) -> Dict[str, dict]:
        notes_data = self.notes.get_data(deep_copy=True)[DeckPartNoteKeys.NOTES.value]
        self.verify_notes_match_note_model_mappings(notes_data)

        csv_data: Dict[str, dict] = {}
        for note in notes_data:
            nm_name = note[DeckPartNoteKeys.NOTE_MODEL.value]
            row = self.note_model_mappings_dict[nm_name].note_model.zip_field_to_data(
                note[DeckPartNoteKeys.FIELDS.value])
            row[CsvKeys.GUID.value] = note[DeckPartNoteKeys.GUID.value]
            row[CsvKeys.TAGS.value] = self.join_tags(note[DeckPartNoteKeys.TAGS.value])

            formatted_row = self.note_model_mappings_dict[nm_name].note_fields_map_to_csv_row(row)

            csv_data.setdefault(row[CsvKeys.GUID.value], formatted_row)

        return csv_data

    def verify_notes_match_note_model_mappings(self, notes):
        note_models_used = {note[DeckPartNoteKeys.NOTE_MODEL.value] for note in notes}
        errors = [TypeError(f"Unknown note model type '{model}' in notes '{self.notes.file_location}. "
                            f"Add {SourceCsvKeys.NOTE_MODEL_MAPPINGS.value} for that model.")
                  for model in note_models_used if model not in self.note_model_mappings_dict.keys()]

        if errors:
            raise Exception(errors)

    def source_to_deck_parts(self):
        notes_data = self.notes_to_deck_parts()
        self.notes.set_data(notes_data)

    def deck_parts_to_source(self):
        csv_data = self.notes_to_source()

        for cfm in self.csv_file_mappings:
            cfm.compile_data()
            cfm.set_relevant_data(csv_data)



import glob
import logging
import pathlib
from pathlib import Path
from typing import List, Dict

from brain_brew.representation.json.json_file import JsonFile
from brain_brew.representation.generic.media_file import MediaFile
from brain_brew.utils import filename_from_full_path, find_all_files_in_directory


class CrowdAnkiExport(JsonFile):
    folder_location: str
    contains_media: bool
    known_media: Dict[str, MediaFile]
    media_loc: str

    def __init__(self, folder_location, read_now=True, data_override=None):
        self.folder_location = folder_location
        if self.folder_location[-1] != "/":
            self.folder_location = self.folder_location + "/"

        json_file_location = self.find_json_file_in_folder()
        self.find_all_media()

        super().__init__(json_file_location, read_now=read_now, data_override=data_override)

    def find_json_file_in_folder(self):
        files = glob.glob(self.folder_location + "*.json")

        if len(files) == 1:
            return files[0]
        elif not files:
            logging.error(f"No json file found in folder '{self.folder_location}'")
            raise FileNotFoundError(self.folder_location)
        else:
            logging.error(f"Multiple json files found in '{self.folder_location}': {files}")
            raise FileExistsError()

    def find_all_media(self):
        self.known_media = {}
        self.media_loc = self.folder_location + "media/"  # TODO: Make media folder if not exists
        self.contains_media = Path(self.media_loc).is_dir()

        if not self.contains_media:
            return

        media_files = find_all_files_in_directory(self.media_loc)

        for full_path in media_files:
            filename = filename_from_full_path(full_path)
            self.known_media.setdefault(filename, MediaFile(full_path, filename))

        logging.info(f"CrowdAnkiExport found {len(self.known_media)} media files in folder")

    def write_file(self, data_override=None):
        super().write_file(data_override)
        for filename, media_file in self.known_media.items():
            media_file.copy_source_to_target()



import copy
from typing import List
from unittest.mock import patch

import pytest

from brain_brew.representation.configuration.csv_file_mapping import CsvFileMappingKeys, CsvFileMappingDerivative, \
    CsvFileMapping
from brain_brew.representation.generic.csv_file import CsvFile
from tests.test_file_manager import get_new_file_manager
from tests.representation.configuration.test_global_config import global_config
from tests.representation.generic.test_csv_file import csv_test1, csv_test2, csv_test3, csv_test1_split1, csv_test1_split2, csv_test2_missing_guids


def setup_csv_fm_config(csv: str, sort_by_columns: List[str] = None, reverse_sort: bool = None,
                        note_model_name: str = None, derivatives: List[dict] = None):
    cfm: dict = {
        CsvFileMappingKeys.CSV_FILE.value: csv
    }
    if sort_by_columns is not None:
        cfm.setdefault(CsvFileMappingKeys.SORT_BY_COLUMNS.value, sort_by_columns)
    if reverse_sort is not None:
        cfm.setdefault(CsvFileMappingKeys.REVERSE_SORT.value, reverse_sort)
    if note_model_name is not None:
        cfm.setdefault(CsvFileMappingKeys.NOTE_MODEL.value, note_model_name)
    if derivatives is not None:
        cfm.setdefault(CsvFileMappingKeys.DERIVATIVES.value, derivatives)

    return cfm


class TestConstructor:
    @pytest.mark.parametrize("read_file_now, note_model_name, csv, sort_by_columns, reverse_sort", [
        (False, "note_model.json", "first.csv", ["guid"], False),
        (True, "model_model.json", "second.csv", ["guid", "note_model_name"], True),
        (False, "note_model-json", "first.csv", ["guid"], False,)
    ])
    def test_runs_without_derivatives(self, global_config, read_file_now, note_model_name, csv,
                                      sort_by_columns, reverse_sort):
        get_new_file_manager()
        config = setup_csv_fm_config(csv, sort_by_columns, reverse_sort, note_model_name=note_model_name)

        def assert_csv(passed_file, read_now):
            assert passed_file == csv
            assert read_now == read_file_now

        with patch.object(CsvFileMappingDerivative, "create_derivative", return_value=None) as mock_derivatives, \
                patch.object(CsvFile, "create", side_effect=assert_csv) as mock_csv:

            csv_fm = CsvFileMapping(config, read_now=read_file_now)

            assert isinstance(csv_fm, CsvFileMapping)
            assert csv_fm.reverse_sort == reverse_sort
            assert csv_fm.sort_by_columns == sort_by_columns
            assert csv_fm.note_model_name == note_model_name

            assert mock_csv.call_count == 1
            assert mock_derivatives.call_count == 0

    @pytest.mark.parametrize("derivatives", [
        [setup_csv_fm_config("test_csv.csv")],
        [setup_csv_fm_config("test_csv.csv"), setup_csv_fm_config("second.csv")],
        [setup_csv_fm_config("a.csv"), setup_csv_fm_config("b.csv"), setup_csv_fm_config("c.csv")],
        [setup_csv_fm_config("a.csv", sort_by_columns=["word", "guid"], reverse_sort=True, note_model_name="d")],
        [setup_csv_fm_config("test_csv.csv", derivatives=[setup_csv_fm_config("der_der.csv")])],
    ])
    def test_runs_with_derivatives(self, global_config, derivatives: list):
        get_new_file_manager()
        config = setup_csv_fm_config("test", [], False, note_model_name="nm", derivatives=derivatives.copy())
        expected_call_count = len(derivatives)

        def assert_der(passed_file, read_now):
            der = derivatives.pop(0)
            assert passed_file == der
            assert read_now is False

        with patch.object(CsvFileMappingDerivative, "create_derivative", side_effect=assert_der) as mock_derivatives, \
                patch.object(CsvFile, "create", return_value=None):

            csv_fm = CsvFileMapping(config, read_now=False)

            assert mock_derivatives.call_count == len(csv_fm.derivatives) == expected_call_count


def csv_fixture_gen(csv_fix):
    with patch.object(CsvFile, "create", return_value=csv_fix):
        csv = CsvFileMapping(setup_csv_fm_config("", note_model_name="Test Model"))
        csv.compile_data()
        return csv


@pytest.fixture()
def csv_file_mapping1(csv_test1):
    return csv_fixture_gen(csv_test1)


@pytest.fixture()
def csv_file_mapping2(csv_test2):
    return csv_fixture_gen(csv_test2)


@pytest.fixture()
def csv_file_mapping3(csv_test3):
    return csv_fixture_gen(csv_test3)


@pytest.fixture()
def csv_file_mapping_split1(csv_test1_split1):
    return csv_fixture_gen(csv_test1_split1)


@pytest.fixture()
def csv_file_mapping_split1(csv_test1_split2):
    return csv_fixture_gen(csv_test1_split2)


@pytest.fixture()
def csv_file_mapping2_missing_guids(csv_test2_missing_guids):
    return csv_fixture_gen(csv_test2_missing_guids)


class TestSetRelevantData:
    def test_no_change(self, csv_file_mapping1: CsvFileMapping, csv_file_mapping_split1: CsvFileMapping):
        assert csv_file_mapping1.data_set_has_changed is False

        previous_data = csv_file_mapping1.compiled_data.copy()
        csv_file_mapping1.set_relevant_data(csv_file_mapping_split1.compiled_data)

        assert previous_data == csv_file_mapping1.compiled_data
        assert csv_file_mapping1.data_set_has_changed is False

    def test_change_but_no_extra(self, csv_file_mapping1: CsvFileMapping, csv_file_mapping2: CsvFileMapping):
        assert csv_file_mapping1.data_set_has_changed is False
        assert len(csv_file_mapping1.compiled_data) == 15

        previous_data = copy.deepcopy(csv_file_mapping1.compiled_data)
        csv_file_mapping1.set_relevant_data(csv_file_mapping2.compiled_data)

        assert previous_data != csv_file_mapping1.compiled_data
        assert csv_file_mapping1.data_set_has_changed is True
        assert len(csv_file_mapping1.compiled_data) == 15

    def test_change_extra_row(self, csv_file_mapping1: CsvFileMapping, csv_file_mapping3: CsvFileMapping):
        assert csv_file_mapping1.data_set_has_changed is False
        assert len(csv_file_mapping1.compiled_data) == 15

        previous_data = copy.deepcopy(csv_file_mapping1.compiled_data.copy())
        csv_file_mapping1.set_relevant_data(csv_file_mapping3.compiled_data)

        assert previous_data != csv_file_mapping1.compiled_data
        assert csv_file_mapping1.data_set_has_changed is True
        assert len(csv_file_mapping1.compiled_data) == 16


class TestCompileData:
    num = 0

    def get_num(self):
        self.num += 1
        return self.num

    def test_when_missing_guids(self, csv_file_mapping2_missing_guids: CsvFileMapping):
        with patch("brain_brew.representation.configuration.csv_file_mapping.generate_anki_guid", wraps=self.get_num) as mock_guid:

            csv_file_mapping2_missing_guids.compile_data()

            assert csv_file_mapping2_missing_guids.data_set_has_changed is True
            assert mock_guid.call_count == 9
            assert list(csv_file_mapping2_missing_guids.compiled_data.keys()) == list(range(1, 10))

#  Tests still to do:
#
#  Top level needs a NoteModel, others do not




from tempfile import NamedTemporaryFile, TemporaryDirectory

import pytest

from brain_brew.constants.deckpart_keys import DeckPartNoteKeys, NoteFlagKeys
from brain_brew.representation.json.json_file import JsonFile
from brain_brew.representation.json.deck_part_header import DeckPartHeader
from brain_brew.representation.json.deck_part_notemodel import DeckPartNoteModel, CANoteModelKeys
from brain_brew.representation.json.deck_part_notes import DeckPartNotes


def debug_write_to_target_json(data, json: JsonFile):
    json.set_data(data)
    json.write_file()


def setup_temp_file_in_folder(file_suffix):
    temp_dir = TemporaryDirectory()
    temp_file = NamedTemporaryFile(suffix=file_suffix, delete=False, dir=temp_dir.name)

    return temp_dir, temp_file


# def get_global_config(note_model_loc: str = "", group_by_note_model=True, extract_shared_tags=False) -> GlobalConfig:
#     config = GlobalConfig.get_instance(override=GlobalConfig({
#         ConfigKeys.DECK_PARTS.value: {
#             "headers": "",
#             "note_models": note_model_loc,
#             "notes": "",
#
#             ConfigKeys.DECK_PARTS_NOTES_STRUCTURE.value: {
#                 NoteFlagKeys.GROUP_BY_NOTE_MODEL.value: group_by_note_model,
#                 NoteFlagKeys.EXTRACT_SHARED_TAGS.value: extract_shared_tags
#             }
#         },
#         ConfigKeys.FLAGS.value: {
#
#         }
#     }))
#
#     return config


# @pytest.fixture()
# def global_config():
#     return GlobalConfig.get_instance(override=GlobalConfig({
#         ConfigKeys.DECK_PARTS.value: {
#             "headers": "",
#             "note_models": "",
#             "notes": "",
#
#             ConfigKeys.DECK_PARTS_NOTES_STRUCTURE.value: {
#                 NoteFlagKeys.GROUP_BY_NOTE_MODEL.value: False,
#                 NoteFlagKeys.EXTRACT_SHARED_TAGS.value: False
#             }
#         },
#         ConfigKeys.FLAGS.value: {
#
#         }
#     }))


def make_deck_part_header_mock(data):
    return DeckPartHeader(
        NamedTemporaryFile(suffix=".json", delete=False).name,
        read_now=False,
        data_override=data
    )


def make_deck_part_note_model_mock(data):
    return DeckPartNoteModel(
        NamedTemporaryFile(suffix=".json", delete=False).name,
        read_now=False,
        data_override=data
    )


def make_deck_part_notes_mock(data):
    return DeckPartNotes(
        NamedTemporaryFile(suffix=".json", delete=False).name,
        data_override=data
    )


@pytest.fixture()
def headers_mock():
    return make_deck_part_header_mock({
        "__type__": "Deck",
        "crowdanki_uuid": "72ac74b8-0077-11ea-959e-d8cb8ac9abf0",
        "deck_config_uuid": "3cc64d85-e410-11e9-960e-d8cb8ac9abf0",
        "name": "LL::1. Vocab"
    })


@pytest.fixture()
def note_models_mock():
    return [
        make_deck_part_note_model_mock({
            "__type__": "NoteModel",
            CANoteModelKeys.ID.value: "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            CANoteModelKeys.NAME.value: "LL Word",
            CANoteModelKeys.FIELDS.value: [
                {"name": "Word"},
                {"name": "OtherWord"}
            ]
        }),
        make_deck_part_note_model_mock({
            "__type__": "NoteModel",
            CANoteModelKeys.ID.value: "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb",
            CANoteModelKeys.NAME.value: "LL Verb",
            CANoteModelKeys.FIELDS.value: [
                {"name": "Word"},
                {"name": "OtherWord"}
            ]
        }),
        make_deck_part_note_model_mock({
            "__type__": "NoteModel",
            CANoteModelKeys.ID.value: "cccccccc-cccc-cccc-cccc-cccccccccccc",
            CANoteModelKeys.NAME.value: "LL Noun",
            CANoteModelKeys.FIELDS.value: [
                {"name": "Word"},
                {"name": "OtherWord"}
            ]
        }),
        make_deck_part_note_model_mock({
            "__type__": "NoteModel",
            CANoteModelKeys.ID.value: "dddddddd-dddd-dddd-dddd-dddddddddddd",
            CANoteModelKeys.NAME.value: "LL Sentence",
            CANoteModelKeys.FIELDS.value: [
                {"name": "Word"},
                {"name": "OtherWord"}
            ]
        }),
        make_deck_part_note_model_mock({
            "__type__": "NoteModel",
            CANoteModelKeys.ID.value: "eeeeeeee-eeee-eeee-eeee-eeeeeeeeeeee",
            CANoteModelKeys.NAME.value: "Cloze",
            CANoteModelKeys.FIELDS.value: [
                {"name": "Word"},
                {"name": "OtherWord"}
            ]
        })
    ]


@pytest.fixture()
def notes_mock():
    return make_deck_part_notes_mock({
        DeckPartNoteKeys.FLAGS.value: {
            NoteFlagKeys.GROUP_BY_NOTE_MODEL.value: True,
            NoteFlagKeys.EXTRACT_SHARED_TAGS.value: False
        },
        "Cloze": {
            "notes": [
                {
                    "fields": [
                        "testers",
                        ""
                    ],
                    "guid": "wpoGpMPjTD",
                    "tags": []
                }
            ]
        },
        "LL Noun": {
            "notes": [
                {
                    "fields": [
                        "banana",
                        "en banan",
                        "banano",
                        "<img src=\"All-About-Bananas-Nutrition-Facts-Health-Benefits-Recipes-and-More-RM-722x406.jpg\">",
                        "",
                        "[sound:pronunciation_da_banan.mp3]",
                        ""
                    ],
                    "guid": "sZGs^rTTEr",
                    "tags": [
                        "LL::Grammar::Noun"
                    ]
                }
            ]
        },
        "LL Sentence": {
            "notes": [
                {
                    "fields": [
                        "I don't understand",
                        "<img src=\".jpg (50)\">",
                        "y",
                        "y",
                        "jeg",
                        "forstår",
                        "ikke",
                        "",
                        "",
                        "mi ne",
                        "komprenas",
                        "",
                        "",
                        ""
                    ],
                    "guid": "x+JWF/d?][",
                    "tags": [
                        "CardType::Sentence"
                    ]
                }
            ]
        },
        "LL Verb": {
            "notes": [
                {
                    "fields": [
                        "to learn",
                        "at lære",
                        "lerni",
                        "<img src=\".jpg (51)\">",
                        "",
                        "[sound:pronunciation_da_lære.mp3]",
                        "",
                        "lærer",
                        "lærte",
                        "har lært"
                    ],
                    "guid": "ONZwlFyCX1",
                    "tags": [
                        "LL::Grammar::Verb"
                    ]
                }
            ]
        },
        "LL Word": {
            "notes": [
                {
                    "fields": [
                        "you",
                        "du",
                        "vi",
                        "<img src=\"You1.gif\">",
                        "Singular, Subjective, 2nd Person",
                        "[sound:pronunciation_da_du.mp3]",
                        ""
                    ],
                    "guid": "uZFR2ToY#3",
                    "tags": [
                        "LL::Grammar::Pronoun"
                    ]
                },
                {
                    "fields": [
                        "want",
                        "test",
                        "",
                        "",
                        "",
                        "",
                        ""
                    ],
                    "guid": "zlEirC^ya/",
                    "tags": [
                        "LL::Grammar::Noun",
                        "LL::Grammar::Verb"
                    ]
                }
            ]
        }
    })



import pytest

from brain_brew.representation.json.crowd_anki_export import CrowdAnkiExport
from brain_brew.representation.json.json_file import JsonFile
from tests.test_files import TestFiles


class TestConstructor:
    @pytest.mark.parametrize("export_name", [
        TestFiles.CrowdAnkiExport.TEST1_FOLDER,
        TestFiles.CrowdAnkiExport.TEST1_FOLDER_WITHOUT_SLASH
    ])
    def test_runs(self, export_name):
        file = CrowdAnkiExport(export_name)

        assert isinstance(file, CrowdAnkiExport)
        assert file.folder_location == TestFiles.CrowdAnkiExport.TEST1_FOLDER
        assert file.file_location == TestFiles.CrowdAnkiExport.TEST1_JSON
        assert len(file.get_data().keys()) == 13


class TestFindJsonFileInFolder:
    def test_no_json_file(self, tmpdir):
        directory = tmpdir.mkdir("test")

        with pytest.raises(FileNotFoundError):
            file = CrowdAnkiExport(directory.strpath)

    def test_too_many_json_files(self, tmpdir):
        directory = tmpdir.mkdir("test")
        file1, file2 = directory.join("file1.json"), directory.join("file2.json")
        file1.write("{}")
        file2.write("{}")

        with pytest.raises(FileExistsError):
            file = CrowdAnkiExport(directory.strpath)


@pytest.fixture()
def ca_export_test1() -> CrowdAnkiExport:
    return CrowdAnkiExport.create(TestFiles.CrowdAnkiExport.TEST1_FOLDER)


@pytest.fixture()
def temp_ca_export_file(tmpdir) -> CrowdAnkiExport:
    folder = tmpdir.mkdir("ca_export")
    file = folder.join("file.json")
    file.write("{}")

    return CrowdAnkiExport(folder.strpath, read_now=False)



from brain_brew.representation.configuration.global_config import GlobalConfig
from brain_brew.representation.json.json_file import JsonFile


class DeckPartHeader(JsonFile):

    @classmethod
    def formatted_file_location(cls, location):
        return cls.get_json_file_location(GlobalConfig.get_instance().deck_parts.headers, location)

    def __init__(self, location, read_now=True, data_override=None):
        super().__init__(
            self.formatted_file_location(location),
            read_now=read_now, data_override=data_override
        )



import pytest

from brain_brew.representation.json.deck_part_notemodel import DeckPartNoteModel, CANoteModelKeys
from brain_brew.representation.json.json_file import JsonFile
from tests.test_files import TestFiles
from tests.representation.configuration.test_global_config import global_config


class TestConstructor:
    @pytest.mark.parametrize("note_model_name", [
        TestFiles.NoteModels.TEST_COMPLETE,
        TestFiles.NoteModels.TEST,
    ])
    def test_run(self, global_config, note_model_name):
        file = DeckPartNoteModel(note_model_name)

        assert isinstance(file, DeckPartNoteModel)
        assert file.file_location == TestFiles.NoteModels.LOC + TestFiles.NoteModels.TEST_COMPLETE
        assert len(file.get_data().keys()) == 13

        assert file.name == "Test Model"
        assert file.id == "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
        assert file.fields == ["Word", "OtherWord"]

    def test_config_location_override(self, global_config):
        loc = "place_for_stuff/"
        filename = "what-a-great-file.json"

        global_config.deck_parts.note_models = loc

        file = DeckPartNoteModel(filename, read_now=False, data_override={
            CANoteModelKeys.NAME.value: "name",
            CANoteModelKeys.ID.value: "id",
            CANoteModelKeys.FIELDS.value: []
        })

        assert file.file_location == loc + filename


@pytest.fixture()
def dp_note_model_test1(global_config) -> DeckPartNoteModel:
    return DeckPartNoteModel.create(TestFiles.NoteModels.TEST_COMPLETE)


def test_read_fields(dp_note_model_test1):
    expected = ["Word", "OtherWord"]
    assert dp_note_model_test1.read_fields() == expected


@pytest.mark.parametrize("fields", ([
    ["word", "unknown field"],
    ["Word", "Unknown Field"]
]))
def test_check_field_overlap(dp_note_model_test1, fields):
    expected_extra = ["unknown field"]
    expected_missing = ["otherword"]

    missing, extra = dp_note_model_test1.check_field_overlap(fields)

    assert missing == expected_missing
    assert extra == expected_extra


class TestZipFieldToData:
    def test_runs_normally(self, dp_note_model_test1):
        data_to_zip = ["Example word", "Another one"]
        expected = {"word": "Example word", "otherword": "Another one"}

        zipped = dp_note_model_test1.zip_field_to_data(data_to_zip)

        assert zipped == expected

    def test_raises_exception_when_list_length_differs(self, dp_note_model_test1):
        with pytest.raises(Exception):
            dp_note_model_test1.zip_field_to_data(["Example"])


@pytest.fixture()
def temp_dp_note_model_file(tmpdir) -> DeckPartNoteModel:
    file = tmpdir.mkdir("note_models").join("file.json")
    file.write("{}")

    return DeckPartNoteModel(file.strpath, read_now=False)



from enum import Enum


class ConfigKeys(Enum):
    AUTHORS = "authors"

    DECK_PARTS = "deck_parts"
    HEADERS = "headers"
    NOTE_MODELS = "note_models"
    NOTES = "notes"
    MEDIA_FILES = "media_files"

    FLAGS = "flags"
    SORT_CASE_INSENSITIVE = "sort_case_insensitive"
    JOIN_VALUES_WITH = "join_values_with"

    DECK_PARTS_NOTES_STRUCTURE = "deck_part_notes_structure"
    NOTE_SORT_ORDER = "note_sort_order"
    REVERSE_SORT = "reverse_sort"



from unittest.mock import patch

from brain_brew.build_tasks.source_crowd_anki import SourceCrowdAnki
from brain_brew.build_tasks.source_csv import SourceCsv
from brain_brew.builder import Builder
from brain_brew.representation.generic.yaml_file import YamlFile
from tests.test_files import TestFiles
from tests.representation.configuration.test_global_config import global_config


class TestConstructor:
    def test_runs(self, global_config):

        with patch.object(SourceCsv, "__init__", return_value=None) as mock_csv, \
                patch.object(SourceCsv, "verify_contents", return_value=None), \
                patch.object(SourceCrowdAnki, "__init__", return_value=None) as mock_ca:

            data = YamlFile.read_file(TestFiles.BuildConfig.ONE_OF_EACH_TYPE)
            builder = Builder(data, global_config, read_now=False)

            assert len(builder.build_tasks) == 2
            assert mock_csv.call_count == 1
            assert mock_ca.call_count == 1



from unittest.mock import patch

import pytest

from brain_brew.representation.generic.csv_file import CsvFile, CsvKeys
from brain_brew.representation.generic.generic_file import GenericFile
from tests.test_files import TestFiles


@pytest.fixture()
def csv_test1():
    return CsvFile(TestFiles.CsvFiles.TEST1)


@pytest.fixture()
def csv_test1_split1():
    return CsvFile(TestFiles.CsvFiles.TEST1_SPLIT1)


@pytest.fixture()
def csv_test1_split2():
    return CsvFile(TestFiles.CsvFiles.TEST1_SPLIT2)


@pytest.fixture()
def csv_test2():
    return CsvFile(TestFiles.CsvFiles.TEST2)


@pytest.fixture()
def csv_test3():
    return CsvFile(TestFiles.CsvFiles.TEST3)


@pytest.fixture()
def csv_test1_not_read_initially_test():
    return CsvFile(TestFiles.CsvFiles.TEST1, read_now=False)


@pytest.fixture()
def csv_test2_missing_guids():
    return CsvFile(TestFiles.CsvFiles.TEST2_MISSING_GUIDS)


@pytest.fixture()
def temp_csv_test1(tmpdir, csv_test1) -> CsvFile:
    file = tmpdir.mkdir("json").join("file.csv")
    file.write("blank")

    return CsvFile.create(file.strpath, data_override=csv_test1.get_data())


class TestConstructor:
    def test_runs(self, csv_test1):
        assert isinstance(csv_test1, CsvFile)
        assert csv_test1.file_location == TestFiles.CsvFiles.TEST1
        assert "guid" in csv_test1.column_headers

    @pytest.mark.parametrize("column_headers", [
        ["first", "second", "third", "etc"],
        ["Word", "OtherWord", "Audio", "Test"],
        ["X", "Field name with spaces", "", "MorphMan_FocusMorph"],
    ])
    def test_data_override(self, column_headers):
        data_override = [{key: "value" for key in column_headers}]
        csv = CsvFile("file", data_override=data_override)

        assert csv.column_headers == column_headers


def test_to_filename_csv():
    expected = "read-this-file.csv"

    assert expected == CsvFile.to_filename_csv("read this file")
    assert expected == CsvFile.to_filename_csv("read-this-file")
    assert expected == CsvFile.to_filename_csv("read-this-file.csv")
    assert expected == CsvFile.to_filename_csv("read          this        file")


class TestReadFile:
    def test_runs(self, csv_test1_not_read_initially_test):
        assert csv_test1_not_read_initially_test.get_data() == []
        assert csv_test1_not_read_initially_test.column_headers == []
        assert csv_test1_not_read_initially_test.file_location == TestFiles.CsvFiles.TEST1
        assert csv_test1_not_read_initially_test.data_state == GenericFile.DataState.NOTHING_READ_OR_SET

        csv_test1_not_read_initially_test.read_file()

        assert len(csv_test1_not_read_initially_test.get_data()) == 15
        assert "guid" in csv_test1_not_read_initially_test.column_headers
        assert csv_test1_not_read_initially_test.data_state == GenericFile.DataState.READ_IN_DATA


class TestWriteFile:
    def test_runs(self, temp_csv_test1: CsvFile, csv_test1: CsvFile):
        temp_csv_test1.write_file()
        temp_csv_test1.read_file()

        assert temp_csv_test1.get_data() == csv_test1.get_data()


class TestSortData:
    @pytest.mark.parametrize("columns, reverse, result_column, expected_results", [
        (["guid"], False, "guid", [(0, "AAAA"), (1, "BBBB"), (2, "CCCC"), (14, "OOOO")]),
        (["guid"], True, "guid", [(14, "AAAA"), (13, "BBBB"), (12, "CCCC"), (0, "OOOO")]),
        (["english"], False, "english", [(0, "banana"), (1, "bird"), (2, "cat"), (14, "you")]),
        (["english"], True, "english", [(14, "banana"), (13, "bird"), (12, "cat"), (0, "you")]),
        (["tags"], False, "tags", [(0, "besttag"), (1, "funny"), (2, "tag2 tag3"), (13, ""), (14, "")]),
        (["esperanto", "english"], False, "esperanto", [(0, "banano"), (1, "birdo"), (6, "vi"), (14, "")]),
        (["esperanto", "guid"], False, "guid", [(7, "BBBB"), (14, "LLLL")]),
    ])
    def test_sort(self, csv_test1: CsvFile, columns, reverse, result_column, expected_results):
        csv_test1.sort_data(columns, reverse)

        sorted_data = csv_test1.get_data()

        for result in expected_results:
            assert sorted_data[result[0]][result_column] == result[1]

    def test_insensitive(self):
        pass



from typing import List
from unittest.mock import patch, MagicMock

import pytest

from brain_brew.build_tasks.source_csv import SourceCsv, SourceCsvKeys
from brain_brew.constants.deckpart_keys import DeckPartNoteKeys
from brain_brew.representation.configuration.csv_file_mapping import CsvFileMapping
from brain_brew.representation.configuration.note_model_mapping import NoteModelMapping
from brain_brew.representation.generic.csv_file import CsvFile
from brain_brew.representation.generic.generic_file import GenericFile
from brain_brew.representation.json.deck_part_notemodel import DeckPartNoteModel
from brain_brew.representation.json.deck_part_notes import DeckPartNotes
from tests.test_files import TestFiles
from tests.representation.configuration.test_global_config import global_config
from tests.representation.generic.test_csv_file import csv_test1, csv_test2, csv_test1_split1, csv_test1_split2
from tests.representation.json.test_deck_part_notes import dp_notes_test1, dp_notes_test2
from tests.representation.json.test_deck_part_note_model import dp_note_model_test1
from tests.representation.configuration.test_note_model_mapping import nmm_test1, setup_nmm_config
from tests.representation.configuration.test_csv_file_mapping import csv_file_mapping1, csv_file_mapping2, \
    setup_csv_fm_config


def setup_source_csv_config(notes: str, nmm: list, csv_mappings: list):
    return {
        SourceCsvKeys.NOTES.value: notes,
        SourceCsvKeys.NOTE_MODEL_MAPPINGS.value: nmm,
        SourceCsvKeys.CSV_MAPPINGS.value: csv_mappings
    }


def get_csv_default(notes: DeckPartNotes, nmm: List[NoteModelMapping], csv_maps: List[CsvFileMapping]) -> SourceCsv:
    csv_source = SourceCsv(setup_source_csv_config("", [], []), read_now=False)

    csv_source.notes = notes
    csv_source.note_model_mappings_dict = {nm_map.note_model.name: nm_map for nm_map in nmm}
    csv_source.csv_file_mappings = csv_maps

    return csv_source


@pytest.fixture()
def csv_source_test1(dp_notes_test1, nmm_test1, csv_file_mapping1) -> SourceCsv:
    return get_csv_default(dp_notes_test1, [nmm_test1], [csv_file_mapping1])


@pytest.fixture()
def csv_source_test1_split1(csv_source_default, csv_test1_split1, dp_notes_test1) -> SourceCsv:
    csv_source_default.csv_file = csv_test1_split1
    csv_source_default.notes = dp_notes_test1
    return csv_source_default


@pytest.fixture()
def csv_source_test1_split2(csv_source_default2, csv_test1_split2, dp_notes_test2) -> SourceCsv:
    csv_source_default2.csv_file = csv_test1_split2
    csv_source_default2.notes = dp_notes_test1
    return csv_source_default2


@pytest.fixture()
def csv_source_test2(dp_notes_test2, nmm_test1, csv_file_mapping2) -> SourceCsv:
    return get_csv_default(dp_notes_test2, [nmm_test1], [csv_file_mapping2])


# @pytest.fixture()
# def temp_csv_source(global_config, tmpdir) -> SourceCsv:
#     file = tmpdir.mkdir("notes").join("file.csv")
#     file.write("test,1,2,3")


class TestConstructor:
    def test_runs(self):
        source_csv = get_csv_default(None, [], [])
        assert isinstance(source_csv, SourceCsv)

    @pytest.mark.parametrize("notes, model, columns, personal_fields, csv_file", [
        ("notes.json", "Test Model", {"a": "b"}, ["extra"], "file.csv")
    ])
    def test_calls_correctly(self, notes, model, columns, personal_fields, csv_file, nmm_test1):
        nmm_config = [setup_nmm_config(model, columns, personal_fields)]
        csv_config = [setup_csv_fm_config(csv_file, note_model_name=model)]

        def assert_csv(config, read_now):
            assert config in csv_config
            assert read_now is False

        def assert_nmm(config, read_now):
            assert config in nmm_config
            assert read_now is False

        def assert_dpn(config, read_now):
            assert config == notes
            assert read_now is False

        with patch.object(CsvFileMapping, "__init__", side_effect=assert_csv), \
                patch.object(NoteModelMapping, "__init__", side_effect=assert_nmm), \
                patch.object(NoteModelMapping, "note_model"), \
                patch.object(DeckPartNotes, "create", side_effect=assert_dpn):

            #nmm_mock.return_value = False

            source_csv = SourceCsv(setup_source_csv_config(
                notes,
                nmm_config,
                csv_config
            ), read_now=False)


    # def test_missing_non_required_columns


class TestSourceToDeckParts:
    def test_runs_first(self, csv_source_test1, dp_notes_test1, csv_source_test2, dp_notes_test2):
        self.run_s2dp(csv_source_test1, dp_notes_test1)
        self.run_s2dp(csv_source_test2, dp_notes_test2)

    @staticmethod
    def run_s2dp(csv_source: SourceCsv, dp_notes: DeckPartNotes):
        def assert_format(notes_data):
            assert notes_data == dp_notes.get_data()[DeckPartNoteKeys.NOTES.value]

        with patch.object(DeckPartNotes, "set_data", side_effect=assert_format) as mock_set_data:
            csv_source.source_to_deck_parts()
            assert mock_set_data.call_count == 1


class TestDeckPartsToSource:
    def test_runs_with_no_change(self, csv_source_test1, csv_test1, csv_source_test2, csv_test2):

        self.run_dpts(csv_source_test1, csv_test1)
        self.run_dpts(csv_source_test2, csv_test2)

    @staticmethod
    def run_dpts(csv_source: SourceCsv, csv_file: CsvFile):
        def assert_format(source_data):
            assert source_data == csv_file.get_data()

        with patch.object(GenericFile, "set_data", side_effect=assert_format) as mock_set_data:
            csv_source.deck_parts_to_source()
            assert csv_source.csv_file_mappings[0].data_set_has_changed is False

            csv_source.csv_file_mappings[0].data_set_has_changed = True
            csv_source.csv_file_mappings[0].write_file_on_close()
            assert mock_set_data.call_count == 1




import logging

import pytest
from unittest.mock import MagicMock

from brain_brew.representation.generic.generic_file import GenericFile
from tests.test_file_manager import get_new_file_manager
from tests.test_files import TestFiles


class TestConstructor:
    def test_runs(self):
        file_location = TestFiles.CsvFiles.TEST1
        file = GenericFile(file_location, read_now=False, data_override=None)

        assert isinstance(file, GenericFile)
        assert file.file_location == file_location
        assert file._data is None

    def test_no_file_found(self):
        file_location = "sdfsdfgdsfsdfsdsdg/sdfsdf/sdfsdf/sdfsd/"

        with pytest.raises(FileNotFoundError):
            GenericFile(file_location, read_now=True, data_override=None)

    def test_override_data(self):
        override_data = {"Test": 1}
        file = GenericFile("", read_now=True, data_override=override_data)

        assert isinstance(file, GenericFile)
        assert file._data == override_data


class TestCreateFileWithFileManager:
    def test_runs(self):
        fm = get_new_file_manager()
        assert len(fm.known_files_dict) == 0

        first = GenericFile.create("test1", read_now=False)

        assert isinstance(first, GenericFile)
        assert len(fm.known_files_dict) == 1
        assert fm.known_files_dict["test1"]

    def test_returns_existing_object(self):
        fm = get_new_file_manager()
        assert len(fm.known_files_dict) == 0

        first = GenericFile.create("test1", read_now=False)
        second = GenericFile.create("test1", read_now=False)

        assert first == second
        assert len(fm.known_files_dict) == 1



import pytest

from brain_brew.build_tasks.build_task_generic import BuildTaskGeneric
from tests.representation.configuration.test_global_config import global_config


class TestSplitTags:
    @pytest.mark.parametrize("str_to_split, expected_result", [
        ("tags1, tags2", ["tags1", "tags2"]),
        ("tags1 tags2", ["tags1", "tags2"]),
        ("tags1; tags2", ["tags1", "tags2"]),
        ("tags1      tags2", ["tags1", "tags2"]),
        ("tags1, tags2, tags3, tags4, tags5, tags6, tags7, tags8, tags9",
            ["tags1", "tags2", "tags3", "tags4", "tags5", "tags6", "tags7", "tags8", "tags9"]),
        ("tags1, tags2; tags3 tags4      tags5,     tags6;    tags7    tags8, tags9",
         ["tags1", "tags2", "tags3", "tags4", "tags5", "tags6", "tags7", "tags8", "tags9"]),
        ("tags1,tags2", ["tags1", "tags2"]),
        ("tags1;tags2", ["tags1", "tags2"]),
        ("tags1,    tags2", ["tags1", "tags2"]),
        ("tags1;    tags2", ["tags1", "tags2"]),
    ])
    def test_runs(self, str_to_split, expected_result):
        assert BuildTaskGeneric.split_tags(str_to_split) == expected_result


class TestJoinTags:
    @pytest.mark.parametrize("join_with, expected_result", [
        (", ", "test, test1, test2")
    ])
    def test_joins(self, global_config, join_with, expected_result):
        list_to_join = ["test", "test1", "test2"]
        global_config.flags.join_values_with = join_with

        assert BuildTaskGeneric.join_tags(list_to_join) == expected_result



from collections import namedtuple
from pathlib import Path

import yaml

from brain_brew.constants.build_config_keys import BuildConfigKeys

ConfigKey = namedtuple("ConfigKey", "required entry_type children")


class YamlFile:
    config_entry: dict
    expected_keys: dict
    subconfig_filter: list

    @staticmethod
    def read_file(src: str):
        if src[-5:] not in [".yaml", ".yml"]:
            src += ".yaml"

        if not Path(src).is_file():
            raise FileNotFoundError(src)

        with open(src, 'r') as yml_file:
            f = yaml.full_load(yml_file)

        return f

    @staticmethod
    def check_config_recursive(expected_keys, keys, parent_key_name=""):
        errors = []

        keys_left_to_check = list(expected_keys.keys())

        for key in keys:
            key_name = f"{parent_key_name}/{key}"

            if key in keys_left_to_check:
                keys_left_to_check.remove(key)

                if expected_keys[key].children:
                    errors = errors + \
                             YamlFile.check_config_recursive(expected_keys[key].children, keys[key], key_name)

                if not isinstance(keys[key], expected_keys[key].entry_type):
                    errors.append(f"Expected '{key_name}' to be of type {expected_keys[key].entry_type}"
                                  f", not {type(keys[key])}")

            else:
                errors.append(f"Unexpected key '{key_name}'")

        for key in keys_left_to_check:
            if expected_keys[key].required:
                errors.append(f"Missing key '{key}'")

        return errors

    def verify_config_entry(self):
        errors = YamlFile.check_config_recursive(self.expected_keys, self.config_entry)
        if errors:
            self.raise_error_in_config(errors)

    def get_config(self, enum_key, otherwise=None):
        if enum_key.value in self.config_entry:
            return self.config_entry[enum_key.value]
        if otherwise is not None:
            return otherwise
        raise KeyError(f"Expected key {enum_key.value}")

    def raise_error_in_config(self, error_message):
        raise KeyError(error_message)

    def setup_config_with_subconfig_replacement(self, config_entry: dict):
        if BuildConfigKeys.SUBCONFIG.value not in config_entry:
            self.config_entry = config_entry
            return

        sub = config_entry[BuildConfigKeys.SUBCONFIG.value]

        clone = config_entry.copy()
        clone.pop(BuildConfigKeys.SUBCONFIG.value)

        def verify_and_read_sub(sub_to_verify: str, keep_only_keys):
            if not isinstance(sub_to_verify, str):
                raise TypeError(f"Unknown type in {BuildConfigKeys.SUBCONFIG.value}")
            data = YamlFile.read_file(sub_to_verify)
            if keep_only_keys is not None:
                return {k: data[k] for k in data if k in keep_only_keys}
            return data

        if isinstance(sub, str):
            replacement_sub: dict = verify_and_read_sub(sub, self.subconfig_filter)
            clone = {**clone, **replacement_sub}
        elif isinstance(sub, list):
            replacement_list = []
            for s in sub:
                replacement_list.append(verify_and_read_sub(s, self.subconfig_filter))
            replacement_sub = {BuildConfigKeys.SUBCONFIG.value: replacement_list}
            clone = {**clone, **replacement_sub}
        else:
            raise TypeError(f"{BuildConfigKeys.SUBCONFIG.value} is the wrong type: {type(sub)}")

        self.config_entry = clone



import pytest

from brain_brew.representation.json.json_file import JsonFile
from tests.test_files import TestFiles


def test_constructor():
    file_location = TestFiles.CrowdAnkiExport.TEST1_JSON
    file = JsonFile(file_location)

    assert isinstance(file, JsonFile)
    assert file.file_location == file_location
    assert len(file.get_data().keys()) == 13


def test_to_filename_json():
    expected = "read-this-file.json"

    assert expected == JsonFile.to_filename_json("read this file")
    assert expected == JsonFile.to_filename_json("read-this-file")
    assert expected == JsonFile.to_filename_json("read-this-file.json")
    assert expected == JsonFile.to_filename_json("read          this        file")


def test_configure_file_location():
    expected = "folder/read-this-file.json"

    assert expected == JsonFile.get_json_file_location("folder/", "read this file")
    assert expected == JsonFile.get_json_file_location("folder/", "read-this-file.json")
    assert expected == JsonFile.get_json_file_location("", "folder/read this file")
    assert expected == JsonFile.get_json_file_location("", "folder/read-this-file.json")


@pytest.fixture()
def temp_json_file(tmpdir) -> JsonFile:
    file = tmpdir.mkdir("json").join("file.json")
    file.write("{}")

    return JsonFile(file.strpath)



import shutil
from enum import Enum


class MediaFile:
    class ManagementType(Enum):
        EXISTS = 0
        OVERRIDDEN = 1
        TO_BE_CLONED = 2

    target_loc: str
    filename: str

    man_type: ManagementType
    source_loc: str

    def __init__(self, target_loc, filename, man_type: ManagementType = ManagementType.EXISTS, source_loc=None):
        self.target_loc = target_loc
        self.filename = filename

        self.man_type = man_type
        self.source_loc = target_loc if source_loc is None else source_loc

    def set_override(self, source_loc):
        self.man_type = MediaFile.ManagementType.OVERRIDDEN
        self.source_loc = source_loc

    def copy_source_to_target(self):
        if self.should_write():
            # TODO: If ManagementType.OVERRIDDEN check if override necessary
            shutil.copy2(self.source_loc, self.target_loc)

    def should_write(self):
        return self.man_type in [MediaFile.ManagementType.OVERRIDDEN, MediaFile.ManagementType.TO_BE_CLONED]



import logging

from brain_brew.argument_reader import BBArgumentReader
from brain_brew.builder import Builder
from brain_brew.file_manager import FileManager
from brain_brew.representation.configuration.global_config import GlobalConfig
from brain_brew.representation.generic.yaml_file import YamlFile


# sys.path.append(os.path.join(os.path.dirname(__file__), "dist"))
# sys.path.append(os.path.dirname(__file__))


def main():
    logging.basicConfig(level=logging.DEBUG)

    # Read in Arguments
    argument_reader = BBArgumentReader()
    builder_file_name, global_config_file, run_reversed = argument_reader.get_parsed()
    builder_config = YamlFile.read_file(builder_file_name)

    # Read in Global Config File
    global_config = GlobalConfig.from_yaml(global_config_file) if global_config_file else GlobalConfig.get_default()
    file_manager = FileManager()

    # Run chosen Builder
    builder = Builder(builder_config, global_config, run_reversed=run_reversed)
    builder.execute()


if __name__ == "__main__":
    main()



import abc


class WritesFile(abc.ABC):
    @abc.abstractmethod
    def write_file_on_close(self):
        pass

    def __init__(self):
        from brain_brew.file_manager import FileManager
        fm = FileManager.get_instance()
        fm.register_write_file_for_end(self)



from dataclasses import dataclass, field

from brain_brew.constants.deckpart_keys import NoteFlagKeys, DeckPartNoteFlags
from brain_brew.constants.global_config_keys import *
from brain_brew.representation.generic.yaml_file import YamlFile, ConfigKey


@dataclass
class DeckPartConfig:
    headers: str
    note_models: str
    notes: str
    media_files: str


class GlobalConfig(YamlFile):
    __instance = None

    @dataclass
    class ConfigFlags:
        note_sort_order: list = field(default_factory=list)
        sort_case_insensitive: bool = False
        reverse_sort: bool = False
        join_values_with: str = " "

    deck_parts: DeckPartConfig
    flags: ConfigFlags
    deck_part_notes_flags: DeckPartNoteFlags
    join_values_with: str

    config_entry = {}
    expected_keys = {
        ConfigKeys.DECK_PARTS.value: ConfigKey(True, dict, {
            ConfigKeys.HEADERS.value: ConfigKey(True, str, None),
            ConfigKeys.NOTE_MODELS.value: ConfigKey(True, str, None),
            ConfigKeys.NOTES.value: ConfigKey(True, str, None),
            ConfigKeys.MEDIA_FILES.value: ConfigKey(True, str, None),

            ConfigKeys.DECK_PARTS_NOTES_STRUCTURE.value: ConfigKey(True, dict, {
                flag.value: ConfigKey(False, bool, None) for flag in NoteFlagKeys
            })
        }),
        ConfigKeys.FLAGS.value: ConfigKey(False, dict, {
            ConfigKeys.NOTE_SORT_ORDER.value: ConfigKey(False, list, None),
            ConfigKeys.SORT_CASE_INSENSITIVE.value: ConfigKey(False, bool, None),
            ConfigKeys.REVERSE_SORT.value: ConfigKey(False, bool, None),
            ConfigKeys.JOIN_VALUES_WITH.value: ConfigKey(False, str, None),
        })
    }
    subconfig_filter = None

    def __init__(self, config_data: dict):
        if GlobalConfig.__instance is None:
            GlobalConfig.__instance = self
        else:
            raise Exception("Multiple GlobalConfigs created")

        self.setup_config_with_subconfig_replacement(config_data)
        self.verify_config_entry()

        dp = self.get_config(ConfigKeys.DECK_PARTS)
        self.deck_parts = DeckPartConfig(
            dp[ConfigKeys.HEADERS.value], dp[ConfigKeys.NOTE_MODELS.value],
            dp[ConfigKeys.NOTES.value], dp[ConfigKeys.MEDIA_FILES.value]
        )

        self.flags = GlobalConfig.ConfigFlags()
        flag_keys = self.get_config(ConfigKeys.FLAGS, {})
        for flag in self.expected_keys[ConfigKeys.FLAGS.value].children.keys():
            if flag in flag_keys:
                setattr(self.flags, flag, flag_keys[flag])

        self.deck_part_notes_flags = DeckPartNoteFlags()
        note_flags = dp[ConfigKeys.DECK_PARTS_NOTES_STRUCTURE.value]
        for flag in note_flags.keys():
            setattr(self.deck_part_notes_flags, flag, note_flags[flag])

    @classmethod
    def get_default(cls):
        global_config_name = "brain_brew_config.yaml"
        return cls.from_yaml(global_config_name)

    @classmethod
    def from_yaml(cls, yaml_file):
        gb_config_yaml = YamlFile.read_file(yaml_file)

        return GlobalConfig(gb_config_yaml)

    @classmethod
    def get_instance(cls):
        return cls.__instance

    @classmethod
    def clear_instance(cls):
        if cls.__instance:
            cls.__instance = None



import logging
from enum import Enum
from typing import Dict, List

from brain_brew.constants.deckpart_keys import DeckPartNoteKeys
from brain_brew.interfaces.verifiable import Verifiable
from brain_brew.interfaces.writes_file import WritesFile
from brain_brew.representation.generic.csv_file import CsvFile, CsvKeys
from brain_brew.representation.generic.generic_file import GenericFile
from brain_brew.representation.generic.yaml_file import YamlFile, ConfigKey
from brain_brew.representation.json.deck_part_notemodel import DeckPartNoteModel
from brain_brew.utils import single_item_to_list, generate_anki_guid, list_of_str_to_lowercase


class CsvFileMappingKeys(Enum):
    CSV_FILE = "csv"
    NOTE_MODEL = "note_model"
    SORT_BY_COLUMNS = "sort_by_columns"
    REVERSE_SORT = "reverse_sort"
    DERIVATIVES = "derivatives"


class CsvFileMappingDerivative(YamlFile):
    config_entry = {}
    expected_keys = {
        CsvFileMappingKeys.CSV_FILE.value: ConfigKey(True, str, None),
        CsvFileMappingKeys.SORT_BY_COLUMNS.value: ConfigKey(False, (list, str), None),
        CsvFileMappingKeys.REVERSE_SORT.value: ConfigKey(False, bool, None),
        CsvFileMappingKeys.DERIVATIVES.value: ConfigKey(False, list, None),

        CsvFileMappingKeys.NOTE_MODEL.value: ConfigKey(False, str, None),  # Optional on Derivatives
    }
    subconfig_filter = None

    csv_file: CsvFile
    compiled_data: Dict[str, dict]

    sort_by_columns: list
    reverse_sort: bool

    note_model_name: str
    derivatives: List['CsvFileMappingDerivative']

    def __init__(self, config_data, read_now=True):
        super().__init__()

        self.setup_config_with_subconfig_replacement(config_data)
        self.verify_config_entry()

        self.csv_file = CsvFile.create(self.get_config(CsvFileMappingKeys.CSV_FILE), read_now=read_now)

        self.sort_by_columns = single_item_to_list(self.get_config(CsvFileMappingKeys.SORT_BY_COLUMNS, []))
        self.reverse_sort = self.get_config(CsvFileMappingKeys.REVERSE_SORT, False)

        self.note_model_name = self.get_config(CsvFileMappingKeys.NOTE_MODEL, "")
        self.note_model_name = None if self.note_model_name == "" else self.note_model_name

        self.derivatives = [CsvFileMappingDerivative.create_derivative(config, read_now=read_now)
                            for config in self.get_config(CsvFileMappingKeys.DERIVATIVES, [])]

    @classmethod
    def create_derivative(cls, config_data, read_now=True):
        return cls(config_data, read_now=read_now)

    def get_available_columns(self):
        return self.csv_file.column_headers + [col for der in self.derivatives for col in der.get_available_columns()]

    def get_used_note_model_names(self) -> List[str]:
        nm = [self.note_model_name] if self.note_model_name is not None else []
        return nm + [name for der in self.derivatives for name in der.get_used_note_model_names()]

    def _build_data_recursive(self) -> List[dict]:
        data_in_progress = self.csv_file.get_data(deep_copy=True)

        new_columns_seen_so_far = self.csv_file.column_headers.copy()
        for der in self.derivatives:
            der_cols = der.get_available_columns()
            overlapping_cols = [col for col in der_cols if col in self.csv_file.column_headers]
            der_cols = [col for col in der_cols if col not in overlapping_cols]

            if not overlapping_cols:
                raise KeyError("No column overlap for derivative")

            column_repeat_errors = [KeyError(f"Derivative column {c} in multiple derivative lines")
                                    for c in der_cols if c in new_columns_seen_so_far]
            if column_repeat_errors:
                raise Exception(column_repeat_errors)
            new_columns_seen_so_far += der_cols

            der_match_errors = []
            for der_row in der._build_data_recursive():
                # Find matching row to pair data with
                found_match = False
                for row in data_in_progress:
                    if all([der_row[c] == row[c] for c in overlapping_cols]):
                        for der_col in der_cols:
                            row[der_col] = der_row[der_col]
                        found_match = True
                        # Set Note Model to matching Derivative Note Model
                        if der.note_model_name is not None:
                            row.setdefault(DeckPartNoteKeys.NOTE_MODEL.value, der.note_model_name)
                        break
                if not found_match:
                    der_match_errors.append(ValueError(f"Cannot match derivative row {der_row} to parent"))

            if der_match_errors:
                raise Exception(der_match_errors)

        return data_in_progress

    def write_to_csv(self, data_to_set):
        self.csv_file.set_data_from_superset(data_to_set)
        self.csv_file.sort_data(self.sort_by_columns, self.reverse_sort)

        for der in self.derivatives:
            der.write_to_csv(data_to_set)


class CsvFileMapping(CsvFileMappingDerivative, Verifiable, WritesFile):
    expected_keys = {
        CsvFileMappingKeys.CSV_FILE.value: ConfigKey(True, str, None),
        CsvFileMappingKeys.SORT_BY_COLUMNS.value: ConfigKey(False, (list, str), None),
        CsvFileMappingKeys.REVERSE_SORT.value: ConfigKey(False, bool, None),
        CsvFileMappingKeys.DERIVATIVES.value: ConfigKey(False, list, None),

        CsvFileMappingKeys.NOTE_MODEL.value: ConfigKey(True, str, None),  # Required on top level
    }

    data_set_has_changed: bool

    def __init__(self, config_data, read_now=True):
        super().__init__(config_data, read_now=read_now)

    def verify_contents(self):
        if self.note_model_name is None:
            raise KeyError(f"Top level Csv Mapping requires key {CsvFileMappingKeys.NOTE_MODEL.value}")

    def compile_data(self):
        self.compiled_data = {}
        self.data_set_has_changed = False

        data_in_progress = self._build_data_recursive()

        # Set Note Model if not already set
        if self.note_model_name is not None:
            for row in data_in_progress:
                row.setdefault(DeckPartNoteKeys.NOTE_MODEL.value, self.note_model_name)

        # Fill in Guid if no Guid
        guids_generated = 0
        for row in data_in_progress:
            guid = row[CsvKeys.GUID.value]
            if not guid:
                guid = row[CsvKeys.GUID.value] = generate_anki_guid()
                guids_generated += 1
            self.compiled_data.setdefault(guid, {key.lower(): row[key] for key in row})

        if guids_generated > 0:
            self.data_set_has_changed = True
            logging.info(f"Generated {guids_generated} guids in {self.csv_file.file_location}")

    def set_relevant_data(self, data_set: Dict[str, dict]):
        unchanged, changed, added = 0, 0, 0
        for guid in data_set:
            if guid in self.compiled_data.keys():
                changed_row = False
                for key in data_set[guid]:
                    if self.compiled_data[guid][key] != data_set[guid][key]:
                        self.compiled_data[guid][key] = data_set[guid][key]
                        changed_row = True
                if changed_row:
                    changed += 1
                else:
                    unchanged += 1
            else:
                added += 1
                self.compiled_data.setdefault(guid, data_set[guid])

        if changed > 0 or added > 0:
            self.data_set_has_changed = True

        logging.info(f"Set {self.csv_file.file_location} data; changed {changed}, "
                     f"added {added}, while {unchanged} were identical")

    def write_file_on_close(self):
        if self.data_set_has_changed:
            self.write_to_csv(list(self.compiled_data.values()))



import pytest

from brain_brew.file_manager import FileManager
from tests.representation.configuration.test_global_config import global_config


def get_new_file_manager():
    FileManager.clear_instance()
    return FileManager()


class TestSingletonConstructor:
    def test_runs(self, global_config):
        fm = get_new_file_manager()
        assert isinstance(fm, FileManager)

    def test_returns_existing_singleton(self):
        fm = get_new_file_manager()
        fm.known_files_dict = {'test': None}
        fm2 = FileManager.get_instance()

        assert fm2.known_files_dict == {'test': None}
        assert fm2 == fm

    def test_raises_error(self):
        with pytest.raises(Exception):
            FileManager()
            FileManager()


class TestFindMediaFiles:
    def test_finds(self):
        fm = get_new_file_manager()

        assert len(fm.known_media_files_dict) == 2



from collections import namedtuple
from enum import Enum

# TODO: Make this a class
BuildTaskEnum = namedtuple("BuildTaskEnum", "key_name source_type task_to_execute reverse_task_to_execute")


class BuildConfigKeys(Enum):
    SUBCONFIG = "subconfig"

    NOTES = "notes"
    HEADERS = "headers"



import json
import re

from brain_brew.representation.generic.generic_file import GenericFile

_encoding = "utf-8"


class JsonFile(GenericFile):
    _data: dict = {}

    def __init__(self, file, read_now=True, data_override=None):
        super(JsonFile, self).__init__(file, read_now=read_now, data_override=data_override)

    def set_data(self, data_override):
        super().set_data(data_override)

    def get_data(self, deep_copy: bool = False) -> dict:
        return super().get_data(deep_copy=deep_copy)

    def pretty_print(self):
        return json.dumps(self._data, indent=4)

    def read_file(self):
        with open(self.file_location, "r", encoding=_encoding) as read_file:
            self._data = json.load(read_file)

    def write_file(self, data_override=None):
        with open(self.file_location, "w", encoding=_encoding) as write_file:
            json.dump(data_override or self._data, write_file, indent=4, sort_keys=False, ensure_ascii=False)

        self.file_exists = True

    @staticmethod
    def to_filename_json(filename: str) -> str:
        converted = re.sub(r'\s+', '-', filename).strip()
        return converted + ".json" if not converted.endswith(".json") else converted

    @staticmethod
    def get_json_file_location(prepend, location):
        return JsonFile.to_filename_json(prepend + location)

    @classmethod
    def formatted_file_location(cls, location):
        return cls.to_filename_json(location)



from argparse import ArgumentParser

import pytest

from brain_brew.argument_reader import BBArgumentReader


@pytest.fixture()
def arg_reader_test1():
    return BBArgumentReader()


def test_constructor(arg_reader_test1):
    assert isinstance(arg_reader_test1, BBArgumentReader)
    assert isinstance(arg_reader_test1, ArgumentParser)


class TestArguments:
    @pytest.mark.parametrize("arguments", [
        (["test_builder.yaml", "--config"]),
        (["test_builder.yaml", "config_file.yaml", "--config"]),
        (["--config", "config_file.yaml"]),
        ([""]),
        ([])
    ])
    def test_broken_arguments(self, arg_reader_test1, arguments):
        with pytest.raises(SystemExit):
            parsed_args = arg_reader_test1.get_parsed(arguments)

    @pytest.mark.parametrize("arguments, builder_file, config_file, run_reversed", [
        (["test_builder.yaml"], "test_builder.yaml", None, False),
        (["test_builder.yaml", "--reversed"], "test_builder.yaml", None, True),
        (["test_builder.yaml", "-r"], "test_builder.yaml", None, True),
        (["test_builder.yaml", "--config", "other_config.yaml"], "test_builder.yaml", "other_config.yaml", False),
        (["test_builder.yaml", "--config", "other_config.yaml", "-r"], "test_builder.yaml", "other_config.yaml", True),
    ])
    def test_correct_arguments(self, arg_reader_test1, arguments, builder_file, config_file, run_reversed):
        parsed_args = arg_reader_test1.parse_args(arguments)

        assert parsed_args.builder_file == builder_file
        assert parsed_args.config_file == config_file
        assert parsed_args.run_reversed == run_reversed



import shutil
from unittest.mock import patch

import pytest

from brain_brew.representation.generic.media_file import MediaFile


@pytest.fixture()
def media_file_test1() -> MediaFile:
    return MediaFile("loc", "name")


class TestConstructor:
    def test_without_override(self):
        loc = "test loc"
        name = "name"

        media_file = MediaFile(loc, name)

        assert isinstance(media_file, MediaFile)
        assert media_file.source_loc == media_file.target_loc == loc
        assert media_file.filename == name
        assert media_file.man_type == MediaFile.ManagementType.EXISTS

    def test_with_override(self):
        source_loc = "test loc"
        target_loc = "other loc"
        name = "name"
        man_type = MediaFile.ManagementType.TO_BE_CLONED

        media_file = MediaFile(target_loc, name, man_type, source_loc)

        assert isinstance(media_file, MediaFile)
        assert media_file.target_loc == target_loc
        assert media_file.source_loc == source_loc
        assert media_file.filename == name
        assert media_file.man_type == man_type


def test_set_override(media_file_test1):
    assert media_file_test1.man_type == MediaFile.ManagementType.EXISTS
    assert media_file_test1.source_loc == media_file_test1.target_loc == "loc"

    media_file_test1.set_override("new loc")

    assert media_file_test1.source_loc == "new loc"
    assert media_file_test1.target_loc == "loc"
    assert media_file_test1.man_type == MediaFile.ManagementType.OVERRIDDEN


@pytest.mark.parametrize("man_type, expected_result", [
    (MediaFile.ManagementType.EXISTS, False),
    (MediaFile.ManagementType.OVERRIDDEN, True),
    (MediaFile.ManagementType.TO_BE_CLONED, True),
])
def test_should_write(media_file_test1, man_type, expected_result):
    media_file_test1.man_type = man_type
    assert media_file_test1.should_write() == expected_result


class TestCopy:
    @pytest.mark.parametrize("should_write_returns, num_calls_to_copy", [
        (True, 1),
        (False, 0)
    ])
    def test_takes_should_write_into_account(self, media_file_test1, should_write_returns, num_calls_to_copy):
        with patch.object(MediaFile, "should_write", return_value=should_write_returns), \
                patch.object(shutil, "copy2") as mock_copy:
            media_file_test1.copy_source_to_target()
            assert mock_copy.call_count == num_calls_to_copy

    def test_copies_file(self, tmpdir):
        source_dir = tmpdir.mkdir("source")
        source = source_dir.join("test.txt")
        source.write("test content")
        assert len(source_dir.listdir()) == 1

        target_dir = tmpdir.mkdir("target")
        target = target_dir.join("test.txt")
        assert len(target_dir.listdir()) == 0

        media_file = MediaFile(target, "test.txt", MediaFile.ManagementType.TO_BE_CLONED, source)
        media_file.copy_source_to_target()

        assert len(target_dir.listdir()) == len(source_dir.listdir()) == 1
        assert target.read() == "test content"



import pytest

from brain_brew.representation.json.deck_part_header import DeckPartHeader
from brain_brew.representation.json.json_file import JsonFile
from tests.test_files import TestFiles
from tests.representation.configuration.test_global_config import global_config


class TestConstructor:
    @pytest.mark.parametrize("header_name", [
        TestFiles.Headers.FIRST,
        TestFiles.Headers.FIRST_COMPLETE,
    ])
    def test_runs(self, header_name, global_config):
        file = DeckPartHeader(header_name)

        assert isinstance(file, DeckPartHeader)
        assert file.file_location == TestFiles.Headers.LOC + TestFiles.Headers.FIRST_COMPLETE
        assert len(file.get_data().keys()) == 10

    def test_config_location_override(self, global_config):
        loc = "place_for_stuff/"
        filename = "what-a-great-file.json"

        global_config.deck_parts.headers = loc

        file = DeckPartHeader(filename, read_now=False, data_override={
            "__type__": "Deck",
            "crowdanki_uuid": "72ac74b8-0077-11ea-959e-d8cb8ac9abf0",
            "deck_config_uuid": "3cc64d85-e410-11e9-960e-d8cb8ac9abf0",
            "name": "LL::1. Vocab"
        })

        assert file.file_location == loc + filename


@pytest.fixture()
def dp_headers_test1(global_config):
    return DeckPartHeader.create(TestFiles.Headers.FIRST_COMPLETE)


@pytest.fixture()
def temp_dp_headers_file(tmpdir) -> DeckPartHeader:
    file = tmpdir.mkdir("headers").join("file.json")
    file.write("{}")

    return DeckPartHeader(file.strpath, read_now=False)



import logging
import re
from typing import List

from brain_brew.file_manager import FileManager
from brain_brew.representation.configuration.global_config import GlobalConfig
from brain_brew.representation.json.json_file import JsonFile
from brain_brew.constants.deckpart_keys import *
from brain_brew.representation.generic.media_file import MediaFile


class CANoteKeys(Enum):
    NOTE_MODEL = "note_model_uuid"
    FLAGS = "flags"
    GUID = "guid"
    TAGS = "tags"
    FIELDS = "fields"


class DeckPartNotes(JsonFile):
    _data: dict = {}
    flags: DeckPartNoteFlags
    referenced_media_files: List[MediaFile]

    global_config: GlobalConfig
    file_manager: FileManager

    @classmethod
    def formatted_file_location(cls, location):
        return cls.get_json_file_location(GlobalConfig.get_instance().deck_parts.notes, location)

    def __init__(self, location, read_now=True, data_override=None):
        self.global_config = GlobalConfig.get_instance()
        self.file_manager = FileManager.get_instance()
        self.flags = DeckPartNoteFlags()
        self.referenced_media_files = []
        super().__init__(
            self.formatted_file_location(location),
            read_now=read_now, data_override=data_override
        )

    def set_data(self, data_override, deck_part_flags_included=False):
        if not deck_part_flags_included:
            data_override = {**self.flags.get_formatted_dict(), **{DeckPartNoteKeys.NOTES.value: data_override}}

        super().set_data(data_override)
        self.interpret_data()

    def write_file(self, data_override=None):
        super().write_file(data_override or self.implement_note_structure())

    def read_file(self):
        super().read_file()
        self.interpret_data()

    def interpret_data(self):
        self.verify_data()

        self.read_note_config()
        self._data = self.remove_notes_structure()
        self.read_note_config()

        self.sort_data(
            self.global_config.flags.note_sort_order,
            self.global_config.flags.reverse_sort
        )

        self.find_all_media_references()

    def read_note_config(self):
        self.flags.group_by_note_model = \
            self._data[DeckPartNoteKeys.FLAGS.value][NoteFlagKeys.GROUP_BY_NOTE_MODEL.value]
        self.flags.extract_shared_tags = \
            self._data[DeckPartNoteKeys.FLAGS.value][NoteFlagKeys.EXTRACT_SHARED_TAGS.value]

    def verify_data(self):
        errors = []
        if DeckPartNoteKeys.FLAGS.value not in self._data:
            errors.append(KeyError(f"Missing '{DeckPartNoteKeys.FLAGS.value}' key in '{self.file_location}'"))
        else:
            for flag in NoteFlagKeys:
                if flag.value not in self._data[DeckPartNoteKeys.FLAGS.value]:
                    errors.append(KeyError(f"Missing '{flag.value}' flag in '{self.file_location}'"))

        if errors:
            raise Exception(errors)

    def get_all_known_note_model_names(self):
        model_set = {note[DeckPartNoteKeys.NOTE_MODEL.value] for note in self._data[DeckPartNoteKeys.NOTES.value]}
        return sorted(list(model_set))

    def implement_note_structure(self):
        """
        :return: the notes structured based on the settings in the Global Config file
        """
        if not self.global_config.deck_part_notes_flags.any_enabled():
            return self._data

        def top_level_structure():
            if self.global_config.deck_part_notes_flags.extract_shared_tags:
                return {DeckPartNoteKeys.SHARED_TAGS.value: [], DeckPartNoteKeys.NOTES.value: []}
            else:
                return {DeckPartNoteKeys.NOTES.value: []}

        if self.global_config.deck_part_notes_flags.group_by_note_model:
            structured_notes: dict = {name: top_level_structure() for name in self.get_all_known_note_model_names()}
            for note in self._data[DeckPartNoteKeys.NOTES.value]:
                for name in structured_notes:
                    if note[DeckPartNoteKeys.NOTE_MODEL.value] == name:
                        structured_notes[name][DeckPartNoteKeys.NOTES.value].append(note)
                        break
                del note[DeckPartNoteKeys.NOTE_MODEL.value]
        else:
            structured_notes = top_level_structure()
            structured_notes[DeckPartNoteKeys.NOTES.value] = self._data

        if self.global_config.deck_part_notes_flags.extract_shared_tags:
            if self.global_config.deck_part_notes_flags.group_by_note_model:
                for name in structured_notes:
                    structured_notes[name][DeckPartNoteKeys.SHARED_TAGS.value] = self.extract_shared_tags(
                        structured_notes[name][DeckPartNoteKeys.NOTES.value]
                    )
            else:
                structured_notes[DeckPartNoteKeys.SHARED_TAGS.value] = self.extract_shared_tags(
                    structured_notes[DeckPartNoteKeys.NOTES.value]
                )

        final_structure = self.global_config.deck_part_notes_flags.get_formatted_dict()
        for key in structured_notes:
            final_structure.setdefault(key, structured_notes[key])

        return final_structure

    @staticmethod
    def extract_shared_tags(notes):
        shared_tags = notes[0][DeckPartNoteKeys.TAGS.value]

        for note in notes:
            shared_tags = [tag for tag in note[DeckPartNoteKeys.TAGS.value] if tag in shared_tags]
            if not shared_tags:
                break

        if shared_tags:
            for note in notes:
                for tag in shared_tags:
                    note[DeckPartNoteKeys.TAGS.value].remove(tag)

        return shared_tags

    def remove_notes_structure(self):
        """
        :return: notes with their global config structure removed
        """
        if not self.flags.any_enabled():
            return self._data

        unstructured_notes = self.get_data(deep_copy=True)

        if self.flags.extract_shared_tags:
            if self.flags.group_by_note_model:
                for group in unstructured_notes:
                    if group[0] != "_":
                        self.spread_out_shared_tags(unstructured_notes[group])
            else:
                self.spread_out_shared_tags(unstructured_notes)

            unstructured_notes[DeckPartNoteKeys.FLAGS.value][NoteFlagKeys.EXTRACT_SHARED_TAGS.value] = False

        if self.flags.group_by_note_model:
            ungrouped_notes = []
            for group in unstructured_notes:
                if group[0] != "_":
                    for note in unstructured_notes[group][DeckPartNoteKeys.NOTES.value]:
                        note[DeckPartNoteKeys.NOTE_MODEL.value] = group
                        ungrouped_notes.append(note)

            top_level_ungrouped = {group: unstructured_notes[group] for group in unstructured_notes if group[0] == "_"}
            top_level_ungrouped.setdefault(DeckPartNoteKeys.NOTES.value, ungrouped_notes)
            unstructured_notes = top_level_ungrouped

            unstructured_notes[DeckPartNoteKeys.FLAGS.value][NoteFlagKeys.GROUP_BY_NOTE_MODEL.value] = False

        return unstructured_notes

    @staticmethod
    def spread_out_shared_tags(top_level):
        shared_tags = top_level[DeckPartNoteKeys.SHARED_TAGS.value]
        if shared_tags:
            for note in top_level[DeckPartNoteKeys.NOTES.value]:
                note[DeckPartNoteKeys.TAGS.value] += shared_tags

        del top_level[DeckPartNoteKeys.SHARED_TAGS.value]

    def sort_data(self, sort_by_keys, reverse_sort):
        sorted_data = self._sort_data(self._data[DeckPartNoteKeys.NOTES.value], sort_by_keys, reverse_sort)

        self._data[DeckPartNoteKeys.NOTES.value] = sorted_data

    def find_all_media_references(self):
        unknown_references = []
        self.referenced_media_files.clear()

        for note in self._data[DeckPartNoteKeys.NOTES.value]:
            for field in note[DeckPartNoteKeys.FIELDS.value]:
                files_found = self.find_media_in_field(field)
                if files_found:
                    for filename in files_found:
                        file = self.file_manager.media_file_if_exists(filename)
                        if file:
                            if file not in self.referenced_media_files:
                                self.referenced_media_files.append(file)
                        else:
                            unknown_references.append(filename)

        if len(unknown_references) > 0:
            logging.error(f"Found {len(unknown_references)} unreferenced media files in {self.file_location}: ["
                          f"{', '.join(unknown_references)}]")

        logging.info(f"Found {len(self.referenced_media_files)} referenced media files")

    @staticmethod
    def find_media_in_field(field_value):
        if not isinstance(field_value, str):
            return []

        images = re.findall(r'<\s*?img.*?src="(.*?)"[^>]*?>', field_value)
        audio = re.findall(r'\[sound:(.*?)\]', field_value)

        return images + audio



from dataclasses import dataclass
from enum import Enum


class DeckPartNoteKeys(Enum):
    NOTES = "notes"
    FIELDS = "fields"
    GUID = "guid"
    NOTE_MODEL = "note_model"
    TAGS = "tags"
    SHARED_TAGS = "_shared_tags"
    FLAGS = "_flags"


class NoteFlagKeys(Enum):
    GROUP_BY_NOTE_MODEL = "group_by_note_model"
    EXTRACT_SHARED_TAGS = "extract_shared_tags"


@dataclass
class DeckPartNoteFlags:
    group_by_note_model: bool = False
    extract_shared_tags: bool = False

    @staticmethod
    def as_formatted_dict(group_by_note_model: bool, extract_shared_tags: bool):
        return {DeckPartNoteKeys.FLAGS.value: {
            NoteFlagKeys.EXTRACT_SHARED_TAGS.value: extract_shared_tags,
            NoteFlagKeys.GROUP_BY_NOTE_MODEL.value: group_by_note_model
        }}

    def get_formatted_dict(self):
        return DeckPartNoteFlags.as_formatted_dict(self.group_by_note_model, self.extract_shared_tags)

    def any_enabled(self):
        return self.group_by_note_model or self.extract_shared_tags



from typing import Dict, List
from unittest.mock import patch

import pytest

from brain_brew.representation.configuration.note_model_mapping import NoteModelMappingKeys, NoteModelMapping, \
    FieldMapping
from brain_brew.representation.generic.csv_file import CsvFile
from brain_brew.representation.json.deck_part_notemodel import DeckPartNoteModel
from tests.representation.configuration.test_global_config import global_config
from tests.test_file_manager import get_new_file_manager
from tests.representation.generic.test_csv_file import csv_test1


def setup_nmm_config(note_model: str, field_mappings: Dict[str, str], personal_fields: List[str]):
    return {
        NoteModelMappingKeys.NOTE_MODEL.value: note_model,
        NoteModelMappingKeys.COLUMNS.value: field_mappings,
        NoteModelMappingKeys.PERSONAL_FIELDS.value: personal_fields
    }


@pytest.fixture()
def nmm_test1(global_config) -> NoteModelMapping:
    get_new_file_manager()
    config = setup_nmm_config(
        "Test Model",
        {
            "guid": "guid",
            "tags": "tags",

            "english": "word",
            "danish": "otherword"
        },
        []
    )
    return NoteModelMapping(config, True)


@pytest.fixture()
def nmm_test_with_personal_fields1(global_config) -> NoteModelMapping:
    get_new_file_manager()
    config = setup_nmm_config(
        "Test Model",
        {
            "guid": "guid",
            "tags": "tags",

            "english": "word",
            "danish": "otherword"
        },
        ["extra", "morph_focus"]
    )
    return NoteModelMapping(config, True)


class TestInit:
    def test_runs(self):
        nmm = NoteModelMapping(setup_nmm_config("test", {}, []), read_now=False)
        assert isinstance(nmm, NoteModelMapping)

    @pytest.mark.parametrize("read_file_now, note_model, personal_fields, columns", [
        (False, "note_model.json", ["x"], {"guid": "guid", "tags": "tags", "english": "word", "danish": "otherword"}),
        (True, "model_model", [], {"guid": "guid", "tags": "tags"}),
        (False, "note_model-json", ["x", "y", "z"], {"guid": "guid", "tags": "tags", "english": "word", "danish": "otherword"})
    ])
    def test_values(self, read_file_now, note_model, personal_fields, columns):
        config = setup_nmm_config(note_model, columns, personal_fields)

        def assert_dp_note_model(passed_file, read_now):
            assert passed_file == note_model
            assert read_now == read_file_now

        with patch.object(DeckPartNoteModel, "create", side_effect=assert_dp_note_model) as mock_nm:

            nmm = NoteModelMapping(config, read_now=read_file_now)

            assert isinstance(nmm, NoteModelMapping)
            assert len(nmm.columns) == len(columns)
            assert len(nmm.personal_fields) == len(personal_fields)

            assert mock_nm.call_count == 1


class TestVerifyContents:
    pass  # TODO


class TestCsvRowNoteFieldConversion:
    @staticmethod
    def get_csv_row(): return {
        "guid": "AAAA",
        "tags": "nice card",

        "english": "what",
        "danish": "hvad"
    }

    @staticmethod
    def get_note_field(): return{
        "guid": "AAAA",
        "tags": "nice card",

        "word": "what",
        "otherword": "hvad",
        "extra": False,
        "morph_focus": False
    }

    def test_csv_row_map_to_note_fields(self, nmm_test_with_personal_fields1):
        assert nmm_test_with_personal_fields1.csv_row_map_to_note_fields(self.get_csv_row()) == self.get_note_field()

    def test_note_fields_map_to_csv_row(self, nmm_test_with_personal_fields1):
        assert nmm_test_with_personal_fields1.note_fields_map_to_csv_row(self.get_note_field()) == self.get_csv_row()


class TestGetRelevantData:
    def test_data_correct(self, nmm_test_with_personal_fields1: NoteModelMapping, csv_test1: CsvFile):
        expected_relevant_columns = ["guid", "english", "danish", "tags"]
        data = csv_test1.get_data()

        for row in data:
            relevant_data = nmm_test_with_personal_fields1.get_relevant_data(row)
            assert len(relevant_data) == 4
            assert list(relevant_data.keys()) == expected_relevant_columns

    def test_data_missing_columns(self, nmm_test_with_personal_fields1: NoteModelMapping, csv_test1: CsvFile):
        row_missing = {
            "guid": "test",
            "english": "test"
        }
        with pytest.raises(Exception) as e:
            relevant_data = nmm_test_with_personal_fields1.get_relevant_data(row_missing)

        errors = e.value.args[0]
        assert len(errors) == 2
        assert isinstance(errors[0], KeyError)
        assert isinstance(errors[1], KeyError)
        assert errors[0].args[0] == "Missing column tags"
        assert errors[1].args[0] == "Missing column danish"


class TestFieldMapping:
    def test_init(self):
        fm = FieldMapping(FieldMapping.FieldMappingType.COLUMN, "Csv_Row", "note_model_field")
        assert isinstance(fm, FieldMapping)
        assert (fm.field_name, fm.value) == ("csv_row", "note_model_field")



import os
import pathlib
import string
import random
import re


def blank_str_if_none(s):
    return '' if s is None else s


def list_of_str_to_lowercase(list_of_strings):
    return [entry.lower() for entry in list_of_strings]


def single_item_to_list(item):
    if isinstance(item, list):
        return item
    return [item]


def filename_from_full_path(full_path):
    return re.findall('[^\\/:*?"<>|\r\n]+$', full_path)[0]


def find_all_files_in_directory(directory, recursive=False):
    found_files = []
    for path, dirs, files in os.walk(directory):
        for file in files:
            found_files.append(os.path.join(path, file))
        if not recursive:
            return found_files
    return found_files


def generate_anki_guid() -> str:
    """Return a base91-encoded 64bit random number."""

    def base62(num: int, extra: str = "") -> str:
        s = string
        table = s.ascii_letters + s.digits + extra
        buf = ""
        while num:
            num, i = divmod(num, len(table))
            buf = table[i] + buf
        return buf

    _base91_extra_chars = "!#$%&()*+,-./:;<=>?@[]^_`{|}~"

    def base91(num: int) -> str:
        # all printable characters minus quotes, backslash and separators
        return base62(num, _base91_extra_chars)

    return base91(random.randint(0, 2 ** 64 - 1))



from enum import Enum
from typing import List

from brain_brew.constants.deckpart_keys import DeckPartNoteKeys
from brain_brew.interfaces.verifiable import Verifiable
from brain_brew.representation.generic.yaml_file import YamlFile, ConfigKey
from brain_brew.representation.json.deck_part_notemodel import DeckPartNoteModel
from brain_brew.utils import list_of_str_to_lowercase


class NoteModelMappingKeys(Enum):
    NOTE_MODEL = "note_model"
    COLUMNS = "csv_columns_to_fields"
    PERSONAL_FIELDS = "personal_fields"


class FieldMapping:
    class FieldMappingType(Enum):
        COLUMN = "column"
        PERSONAL_FIELD = "personal_field"
        DEFAULT = "default"

        @classmethod
        def values(cls):
            return set(it.value for it in cls)

    type: FieldMappingType
    value: str
    field_name: str

    def __init__(self, field_type: FieldMappingType, field_name: str, value: str):
        self.type = field_type
        self.field_name = field_name.lower()

        if self.type == FieldMapping.FieldMappingType.COLUMN:
            self.value = value.lower()
        else:
            self.value = value


class NoteModelMapping(YamlFile, Verifiable):
    config_entry = {}
    expected_keys = {
        NoteModelMappingKeys.NOTE_MODEL.value: ConfigKey(True, str, None),
        NoteModelMappingKeys.COLUMNS.value: ConfigKey(True, dict, None),
        NoteModelMappingKeys.PERSONAL_FIELDS.value: ConfigKey(False, list, None),
    }
    subconfig_filter = None

    note_model: DeckPartNoteModel = None
    columns: List[FieldMapping]
    personal_fields: List[FieldMapping]

    required_fields_definitions = [DeckPartNoteKeys.GUID.value, DeckPartNoteKeys.TAGS.value]

    def __init__(self, config_data: dict, read_now=True):
        self.setup_config_with_subconfig_replacement(config_data)
        self.verify_config_entry()

        columns = self.get_config(NoteModelMappingKeys.COLUMNS)
        personal_fields = self.get_config(NoteModelMappingKeys.PERSONAL_FIELDS, [])

        self.columns = [FieldMapping(
            field_type=FieldMapping.FieldMappingType.COLUMN,
            field_name=field,
            value=columns[field]) for field in columns]

        self.personal_fields = [FieldMapping(
            field_type=FieldMapping.FieldMappingType.PERSONAL_FIELD,
            field_name=field,
            value="") for field in personal_fields]

        self.note_model = DeckPartNoteModel.create(self.get_config(NoteModelMappingKeys.NOTE_MODEL), read_now=read_now)

    def verify_contents(self):
        errors = []

        # Check for Required Fields
        missing = []
        for req in self.required_fields_definitions:
            if req not in [field.value for field in self.columns]:
                missing.append(req)

        if missing:
            errors.append(KeyError(f"""Note model "{self.note_model.name}" to Csv config error: \
                               Definitions for fields {missing} are required."""))

        # Check Fields Align with Note Type
        missing, extra = self.note_model.check_field_overlap(
            [field.value for field in self.columns if field.value not in self.required_fields_definitions]
        )
        missing = [m for m in missing if m not in [field.field_name for field in self.personal_fields]]

        if missing or extra:
            raise KeyError(
                f"""Note model "{self.note_model.name}" to Csv config error. It expected {self.note_model.fields} \
                    but was missing: {missing}, and got extra: {extra} """)

        if errors:
            raise Exception(errors)

    def csv_row_map_to_note_fields(self, row: dict) -> dict:
        relevant_row_data = self.get_relevant_data(row)

        for pf in self.personal_fields:  # Add in Personal Fields
            relevant_row_data.setdefault(pf.field_name, False)
        for column in self.columns:  # Rename from Csv Column to Note Type Field
            relevant_row_data[column.value] = relevant_row_data.pop(column.field_name)

        # TODO: Insert FieldMappings with Default values

        return relevant_row_data

    def csv_headers_map_to_note_fields(self, row: list) -> list:
        return list(self.csv_row_map_to_note_fields({row_name: "" for row_name in row}).keys())

    def note_fields_map_to_csv_row(self, row):
        for column in self.columns:  # Rename from Note Type Field to Csv Column
            row[column.field_name] = row.pop(column.value)

        for pf in self.personal_fields:  # Remove Personal Fields
            del row[pf.field_name]

        relevant_row_data = self.get_relevant_data(row)

        return relevant_row_data

    def get_relevant_data(self, row):
        relevant_columns = [field.field_name for field in self.columns]
        if not relevant_columns:
            return []

        cols = row.keys()

        errors = [KeyError(f"Missing column {rel_col}") for rel_col in relevant_columns if rel_col not in cols]
        if errors:
            raise Exception(errors)

        irrelevant_columns = [column for column in cols if column not in relevant_columns]
        if not irrelevant_columns:
            return row

        relevant_data = {key: row[key] for key in row if key not in irrelevant_columns}

        return relevant_data

    def field_values_in_note_model_order(self, fields_from_csv):
        return [fields_from_csv[field] for field in self.note_model.fields_lowercase]



import pytest

from brain_brew.representation.json.json_file import JsonFile
from brain_brew.representation.json.deck_part_notes import DeckPartNotes, DeckPartNoteKeys
from tests.test_files import TestFiles
from tests.test_helpers import note_models_mock
from tests.representation.configuration.test_global_config import global_config


class TestConstructor:
    @pytest.mark.parametrize("file_to_read", [
        TestFiles.NoteFiles.TEST1_WITH_SHARED_TAGS_EMPTY_AND_GROUPING,
        TestFiles.NoteFiles.TEST1_WITH_SHARED_TAGS,
        TestFiles.NoteFiles.TEST1_WITH_GROUPING,
        TestFiles.NoteFiles.TEST1_NO_GROUPING_OR_SHARED_TAGS,
    ])
    def test_runs_and_unstructures_data(self, file_to_read, global_config, dp_notes_test1):
        expected_result = dp_notes_test1.get_data()
        notes = DeckPartNotes(file_to_read)

        assert isinstance(notes, DeckPartNotes)
        assert notes.get_data() == expected_result


@pytest.fixture()
def dp_notes_test1(global_config) -> DeckPartNotes:
    return DeckPartNotes.create(TestFiles.NoteFiles.TEST1_WITH_SHARED_TAGS_EMPTY_AND_GROUPING)


@pytest.fixture()
def dp_notes_test2(global_config) -> DeckPartNotes:
    return DeckPartNotes.create(TestFiles.NoteFiles.TEST2_WITH_SHARED_TAGS_AND_GROUPING)


@pytest.fixture()
def temp_dp_notes_file(tmpdir) -> DeckPartNotes:
    file = tmpdir.mkdir("notes").join("file.json")
    file.write("{}")

    return DeckPartNotes(file.strpath, read_now=False)


class TestSortData:
    @pytest.mark.parametrize("keys, reverse, result_column, expected_results", [
        (["guid"], False, "guid", [(0, "AAAA"), (1, "BBBB"), (2, "CCCC"), (14, "OOOO")]),
        (["guid"], True, "guid", [(14, "AAAA"), (13, "BBBB"), (12, "CCCC"), (0, "OOOO")]),
        # (["word"], False, "word", [(0, "banana"), (1, "bird"), (2, "cat"), (14, "you")]),
        # (["word"], True, "word", [(14, "banana"), (13, "bird"), (12, "cat"), (0, "you")]),
        # (["tags"], False, "tags", [(0, "besttag"), (1, "funny"), (2, "tag2, tag3"), (13, ""), (14, "")]),
    ])
    def test_sort(self, dp_notes_test1: DeckPartNotes, keys, reverse, result_column, expected_results):
        dp_notes_test1.sort_data(
            keys, reverse
        )

        sorted_data = dp_notes_test1.get_data()[DeckPartNoteKeys.NOTES.value]

        for result in expected_results:
            assert sorted_data[result[0]][result_column] == result[1]

    def test_insensitive(self):
        pass


class TestFindMedia:
    @pytest.mark.parametrize("field_value, expected_results", [
        (r'<img src="image.png">', ["image.png"]),
        (r'<img src="image.png" >', ["image.png"]),
        (r'< img src="image.png">', ["image.png"]),
        (r'<     img src="image.png">', ["image.png"]),
        (r'<img class="myamainzingclass" src="image.png" >', ["image.png"]),
        (r'<img class=noquotesclass src="image.png" >', ["image.png"]),
        (r'<img src="image.png" class=classattheendforsomereason>', ["image.png"]),
        (r'words in the field <img src="image.png" > end other stuff', ["image.png"]),
        (r'<img src="ug-map-saint_barthelemy.png" />', ["ug-map-saint_barthelemy.png"]),
        (r'<img src="ug-map-saint_barthelemy.png" /><img src="image.png">',
            ["ug-map-saint_barthelemy.png", "image.png"]),
        (r'[sound:test.mp3]', ["test.mp3"]),
        (r'[sound:test.mp3][sound:othersound.mp3]', ["test.mp3", "othersound.mp3"]),
        (r'[sound:test.mp3]                       [sound:othersound.mp3]', ["test.mp3", "othersound.mp3"]),
        (r'words in the field [sound:test.mp3] other stuff too [sound:othersound.mp3] end', ["test.mp3", "othersound.mp3"]),
    ])
    def test_find_media_in_field(self, field_value, expected_results):
        assert DeckPartNotes.find_media_in_field(field_value) == expected_results


# def test_set_data_from_override():
#     assert False
#
#
# def test_write_file():
#     assert False
#
#
# def test_read_file():
#     assert False
#
#
# def test_interpret_data():
#     assert False
#
#
# def test_read_note_config():
#     assert False
#
#
# def test_implement_note_structure():
#     assert False
#
#
# def test_remove_notes_structure():
#     assert False



import sys
from argparse import ArgumentParser


class BBArgumentReader(ArgumentParser):
    def __init__(self):
        super().__init__(
            description='Manage Flashcards by Transforming them to various types'
        )

        self._set_parser_arguments()

    def _set_parser_arguments(self):
        self.add_argument(
            "builder_file",
            metavar="builder file",
            type=str,
            help="Run a builder file"
        )
        self.add_argument(
            "--config", "--global-config", "-c",
            action="store",
            dest="config_file",
            default=None,
            type=str,
            help="Global config file to use"
        )

        self.add_argument(
            "-r", "--reversed",
            action="store_true",
            dest="run_reversed",
            default=False,
            help="Run the builder file in reverse"
        )

    def get_parsed(self, override_args=None):
        parsed_args = self.parse_args(args=override_args)

        # Required
        builder = self.error_if_blank(parsed_args.builder_file)

        # Optional
        config_file = parsed_args.config_file
        run_reversed = parsed_args.run_reversed

        return builder, config_file, run_reversed

    def error_if_blank(self, arg):
        if arg == "" or arg is None:
            self.error("Required argument missing")
        return arg

    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)



from enum import Enum
from typing import List

from brain_brew.utils import list_of_str_to_lowercase
from brain_brew.representation.configuration.global_config import GlobalConfig
from brain_brew.representation.json.json_file import JsonFile


class CANoteModelKeys(Enum):
    ID = "crowdanki_uuid"
    NAME = "name"
    FIELDS = "flds"


class DeckPartNoteModel(JsonFile):
    name: str
    id: str
    fields: List[str]

    @property
    def fields_lowercase(self):
        return list_of_str_to_lowercase(self.fields)

    @classmethod
    def formatted_file_location(cls, location):
        return cls.get_json_file_location(GlobalConfig.get_instance().deck_parts.note_models, location)

    def __init__(self, location, read_now=True, data_override=None):
        super().__init__(
            self.formatted_file_location(location),
            read_now=read_now, data_override=data_override
        )

        if read_now or data_override:
            self.name = self._data[CANoteModelKeys.NAME.value]
            self.id = self._data[CANoteModelKeys.ID.value]
            self.fields = self.read_fields()

    def read_fields(self) -> List[str]:
        return [field[CANoteModelKeys.NAME.value] for field in self._data[CANoteModelKeys.FIELDS.value]]

    def check_field_overlap(self, fields_to_check: List[str]):
        fields_to_check = list_of_str_to_lowercase(fields_to_check)
        lower_fields = self.fields_lowercase

        missing = [field for field in lower_fields if field not in fields_to_check]
        extra = [field for field in fields_to_check if field not in lower_fields]

        return missing, extra

    def zip_field_to_data(self, data: List[str]) -> dict:
        if len(self.fields) != len(data):
            raise Exception(f"Data of length {len(data)} cannot map to fields of length {len(self.fields_lowercase)}")
        return dict(zip(self.fields_lowercase, data))



import re

from brain_brew.representation.configuration.global_config import GlobalConfig


class BuildTaskGeneric:
    @staticmethod
    def split_tags(tags_value: str) -> list:
        split = [entry.strip() for entry in re.split(';\s*|,\s*|\s+', tags_value)]
        while "" in split:
            split.remove("")
        return split

    @staticmethod
    def join_tags(tags_list: list) -> str:
        return GlobalConfig.get_instance().flags.join_values_with.join(tags_list)



from enum import Enum

from brain_brew.constants.build_config_keys import BuildTaskEnum
from brain_brew.build_tasks.source_crowd_anki import SourceCrowdAnki
from brain_brew.build_tasks.source_csv import SourceCsv
from brain_brew.file_manager import FileManager
from brain_brew.interfaces.verifiable import Verifiable
from brain_brew.utils import single_item_to_list
from brain_brew.representation.configuration.global_config import GlobalConfig
from brain_brew.representation.generic.yaml_file import YamlFile, ConfigKey


class BuilderKeys(Enum):
    TASKS = "tasks"
    REVERSE_RUN_DIRECTION = "reverse"


class Builder(YamlFile):
    config_entry = {}
    expected_keys = {
        BuilderKeys.TASKS.value: ConfigKey(True, list, None),
        BuilderKeys.REVERSE_RUN_DIRECTION.value: ConfigKey(False, bool, None)
    }
    subconfig_filter = None

    global_config: GlobalConfig

    BUILD_TASK_DEFINITIONS: dict
    KNOWN_BUILD_TASK_CLASSES = [SourceCrowdAnki, SourceCsv]

    build_tasks = []
    file_manager: FileManager

    def __init__(self, config_data, global_config, run_reversed=False, read_now=True):
        self.file_manager = FileManager.get_instance()

        self.BUILD_TASK_DEFINITIONS = {build_task.key_name: build_task
                                       for source in self.KNOWN_BUILD_TASK_CLASSES
                                       for build_task in source.get_build_keys()
                                       }

        self.setup_config_with_subconfig_replacement(config_data)
        self.verify_config_entry()

        self.global_config = global_config

        self.reverse_run_direction = run_reversed or self.get_config(BuilderKeys.REVERSE_RUN_DIRECTION, False)

        # Tasks
        for key in self.config_entry[BuilderKeys.TASKS.value]:
            tasks = single_item_to_list(key)
            for task in tasks:
                task_keys = list(task.keys())
                if len(task_keys) != 1:
                    raise KeyError(f"Task should only contain 1 entry, but contains {task_keys} instead", task)

                if task_keys[0] in self.BUILD_TASK_DEFINITIONS.keys():
                    definition: BuildTaskEnum = self.BUILD_TASK_DEFINITIONS[task_keys[0]]
                    source = definition.source_type(task[task_keys[0]], read_now)
                    if self.reverse_run_direction:
                        self.build_tasks.append((source, definition.reverse_task_to_execute))
                    else:
                        self.build_tasks.append((source, definition.task_to_execute))
                else:
                    raise KeyError(f"Unknown key {key}")  # TODO: check this first on all and return all errors

        if self.reverse_run_direction:
            self.build_tasks = list(reversed(self.build_tasks))

        # Verify tasks
        for source, task_to_execute in self.build_tasks:
            if isinstance(source, Verifiable):
                source.verify_contents()

    def execute(self):
        for (source, task_to_execute) in self.build_tasks:
            getattr(source, task_to_execute)()

        self.file_manager.write_to_all()



from unittest.mock import patch

import pytest

from brain_brew.constants.deckpart_keys import NoteFlagKeys
from brain_brew.constants.global_config_keys import ConfigKeys

from brain_brew.representation.configuration.global_config import GlobalConfig
from tests.test_files import TestFiles


class TestSingletonConstructor:
    def test_runs(self):
        fm = GlobalConfig.get_instance()
        assert isinstance(fm, GlobalConfig)

    def test_returns_existing_singleton(self):
        fm = GlobalConfig.get_instance()
        fm.known_files_dict = {'test': None}
        fm2 = GlobalConfig.get_instance()

        assert fm2.known_files_dict == {'test': None}
        assert fm2 == fm

    def test_raises_error(self):
        with pytest.raises(Exception):
            GlobalConfig({})
            GlobalConfig({})


@pytest.fixture()
def global_config():
    GlobalConfig.clear_instance()
    return GlobalConfig({
        ConfigKeys.DECK_PARTS.value: {
            ConfigKeys.HEADERS.value: TestFiles.Headers.LOC,
            ConfigKeys.NOTE_MODELS.value: TestFiles.NoteModels.LOC,
            ConfigKeys.NOTES.value: TestFiles.NoteFiles.LOC,
            ConfigKeys.MEDIA_FILES.value: TestFiles.MediaFiles.LOC,

            ConfigKeys.DECK_PARTS_NOTES_STRUCTURE.value: {
                NoteFlagKeys.GROUP_BY_NOTE_MODEL.value: False,
                NoteFlagKeys.EXTRACT_SHARED_TAGS.value: False
            }
        },
        ConfigKeys.FLAGS.value: {

        }
    })



import glob
import logging
import pathlib
from typing import Dict, List, Union

from brain_brew.interfaces.writes_file import WritesFile
from brain_brew.representation.configuration.global_config import GlobalConfig
from brain_brew.representation.generic.generic_file import GenericFile
from brain_brew.representation.generic.media_file import MediaFile
from brain_brew.utils import filename_from_full_path, find_all_files_in_directory


class FileManager:
    __instance = None
    global_config: GlobalConfig

    known_files_dict: Dict[str, GenericFile]
    known_media_files_dict: Dict[str, MediaFile]

    write_files_at_end: List[WritesFile]

    def __init__(self):
        if FileManager.__instance is None:
            FileManager.__instance = self
        else:
            raise Exception("Multiple FileManagers created")

        self.global_config = GlobalConfig.get_instance()

        self.known_files_dict = {}
        self.write_files_at_end = []

        self.find_all_deck_part_media_files()

    @staticmethod
    def get_instance() -> 'FileManager':
        return FileManager.__instance

    @staticmethod
    def clear_instance():
        if FileManager.__instance:
            FileManager.__instance = None

    def file_if_exists(self, file_location) -> Union[GenericFile, None]:
        if file_location in self.known_files_dict.keys():
            return self.known_files_dict[file_location]
        return None

    def register_file(self, full_path, file):
        if full_path in self.known_files_dict:
            raise FileExistsError("File already known to FileManager, cannot be registered twice")
        self.known_files_dict.setdefault(full_path, file)

    def media_file_if_exists(self, filename) -> Union[MediaFile, None]:
        if filename in self.known_media_files_dict.keys():
            return self.known_media_files_dict[filename]
        return None
    
    def _register_media_file(self, file: MediaFile):
        if file.filename not in self.known_media_files_dict:
            self.known_media_files_dict.setdefault(file.filename, file)
        else:
            logging.error(f"Duplicate media file '{file.filename}' in both '{file.source_loc}'"
                          f" and '{self.known_media_files_dict[file.filename].source_loc}'")

    def register_write_file_for_end(self, file: WritesFile):
        self.write_files_at_end.append(file)

    def new_media_file(self, filename, source_loc):
        self._register_media_file(MediaFile(self.global_config.deck_parts.media_files + filename,
                                            filename, MediaFile.ManagementType.TO_BE_CLONED, source_loc))

    def find_all_deck_part_media_files(self):
        self.known_media_files_dict = {}

        for full_path in find_all_files_in_directory(self.global_config.deck_parts.media_files, recursive=True):
            filename = filename_from_full_path(full_path)
            self._register_media_file(MediaFile(full_path, filename))

        logging.debug(f"Media files found: {len(self.known_media_files_dict)}")

    def write_to_all(self):
        files_to_create = []
        for location, file in self.known_files_dict.items():
            if not file.file_exists:
                files_to_create.append(location)

        # logging.info(f"Will create {len(files_to_create)} new files: ", files_to_create)

        for write_file in self.write_files_at_end:
            write_file.write_file_on_close()

        for location, file in self.known_files_dict.items():
            if file.data_state == GenericFile.DataState.DATA_SET:
                logging.info(f"Wrote to {file.file_location}")
                file.write_file()
                file.data_state = GenericFile.DataState.READ_IN_DATA

        for filename, media_file in self.known_media_files_dict.items():
            media_file.copy_source_to_target()



from unittest.mock import patch

import pytest

from brain_brew.constants.build_config_keys import BuildConfigKeys
from brain_brew.build_tasks.source_crowd_anki import SourceCrowdAnki, CrowdAnkiKeys
from brain_brew.representation.json.crowd_anki_export import CrowdAnkiExport
from brain_brew.representation.json.deck_part_header import DeckPartHeader
from brain_brew.representation.json.deck_part_notemodel import DeckPartNoteModel
from brain_brew.representation.json.deck_part_notes import DeckPartNotes
from tests.test_files import TestFiles
from tests.representation.json.test_crowd_anki_export import ca_export_test1, temp_ca_export_file
from tests.representation.json.test_deck_part_header import dp_headers_test1, temp_dp_headers_file
from tests.representation.json.test_deck_part_notes import dp_notes_test1, temp_dp_notes_file
from tests.representation.json.test_deck_part_note_model import dp_note_model_test1, temp_dp_note_model_file
from tests.representation.json.test_json_file import temp_json_file
from tests.representation.configuration.test_global_config import global_config


def setup_ca_config(file, media, useless_note_keys, notes, headers):
    return {
        CrowdAnkiKeys.FILE.value: file,
        CrowdAnkiKeys.MEDIA.value: media,
        CrowdAnkiKeys.USELESS_NOTE_KEYS.value: useless_note_keys,
        BuildConfigKeys.NOTES.value: notes,
        BuildConfigKeys.HEADERS.value: headers
    }


class TestConstructor:
    @pytest.mark.parametrize("file, media, useless_note_keys, notes, headers, read_file_now", [
        ("test", False, {}, "test.json", "header.json", False),
        ("export1", True, {}, "test.json", "header.json", False),
        ("json.json", False, {}, "test.json", "", True),
        ("", False, {"__type__": "Note", "data": None, "flags": 0}, "test.json", "header.json", False)
    ])
    def test_runs(self, file, media, useless_note_keys, notes, headers, read_file_now, global_config):
        config = setup_ca_config(file, media, useless_note_keys, notes, headers)

        def assert_dp_header(passed_file, read_now):
            assert passed_file == headers
            assert read_now == read_file_now

        def assert_dp_notes(passed_file, read_now):
            assert passed_file == notes
            assert read_now == read_file_now

        def assert_ca_export(passed_file, read_now):
            assert passed_file == file
            assert read_now == read_file_now

        with patch.object(DeckPartHeader, "create", side_effect=assert_dp_header) as mock_header, \
                patch.object(DeckPartNotes, "create", side_effect=assert_dp_notes) as mock_notes, \
                patch.object(CrowdAnkiExport, "create", side_effect=assert_ca_export) as ca_export:

            source = SourceCrowdAnki(config, read_now=read_file_now)

            assert isinstance(source, SourceCrowdAnki)
            assert source.should_handle_media == media
            assert source.useless_note_keys == useless_note_keys

            assert mock_header.call_count == 1
            assert mock_notes.call_count == 1
            assert ca_export.call_count == 1


@pytest.fixture()
def source_crowd_anki_test1(global_config) -> SourceCrowdAnki:
    with patch.object(DeckPartHeader, "create", return_value=None) as mock_header, \
         patch.object(DeckPartNotes, "create", return_value=None) as mock_notes, \
         patch.object(CrowdAnkiExport, "create", return_value=None) as mock_ca_export:

        source = SourceCrowdAnki(
            setup_ca_config("", False, {"__type__": "Note", "data": None, "flags": 0}, "", "")
        )

        # source.notes = dp_notes_test1
        # source.headers = dp_headers_test1
        # source.crowd_anki_export = ca_export_test1

    return source


class TestSourceToDeckParts:
    def test_runs(self, source_crowd_anki_test1: SourceCrowdAnki, ca_export_test1,
                  temp_dp_note_model_file, temp_dp_headers_file, temp_dp_notes_file,
                  dp_note_model_test1, dp_headers_test1, dp_notes_test1):

        # CrowdAnki Export it will use to write to the DeckParts
        source_crowd_anki_test1.crowd_anki_export = ca_export_test1

        # DeckParts to be written to (+ the NoteModel below)
        source_crowd_anki_test1.headers = temp_dp_headers_file
        source_crowd_anki_test1.notes = temp_dp_notes_file

        def assert_note_model(name, data_override):
            assert data_override == dp_note_model_test1.get_data()
            return dp_note_model_test1

        with patch.object(DeckPartNoteModel, "create", side_effect=assert_note_model) as mock_nm:
            source_crowd_anki_test1.source_to_deck_parts()

            assert source_crowd_anki_test1.headers.get_data() == dp_headers_test1.get_data()
            assert source_crowd_anki_test1.notes.get_data() == dp_notes_test1.get_data()

            assert mock_nm.call_count == 1


class TestDeckPartsToSource:
    def test_runs(self, source_crowd_anki_test1: SourceCrowdAnki, temp_ca_export_file,
                  ca_export_test1, dp_notes_test1, dp_headers_test1):
        source_crowd_anki_test1.crowd_anki_export = temp_ca_export_file  # File to write result to

        # DeckParts it will use (+ dp_note_model_test1, but it reads that in as a file)
        source_crowd_anki_test1.headers = dp_headers_test1
        source_crowd_anki_test1.notes = dp_notes_test1

        source_crowd_anki_test1.deck_parts_to_source()  # Where the magic happens

        assert temp_ca_export_file.get_data() == ca_export_test1.get_data()



import copy
from enum import Enum
from pathlib import Path

from brain_brew.representation.configuration.global_config import GlobalConfig


class GenericFile:
    class DataState(Enum):
        NOTHING_READ_OR_SET = 0
        READ_IN_DATA = 1
        DATA_SET = 2

    _data = None
    file_location: str

    file_exists: bool
    data_state: DataState = DataState.NOTHING_READ_OR_SET

    def __init__(self, file, read_now, data_override):
        self.file_location = file

        self.file_exists = Path(file).is_file()

        if data_override:
            self.data_state = GenericFile.DataState.DATA_SET
            self.set_data(data_override)
        elif read_now:
            if not self.file_exists:
                raise FileNotFoundError(file)
            self.data_state = GenericFile.DataState.READ_IN_DATA
            self.read_file()

    def set_data(self, data_override):
        self.data_state = GenericFile.DataState.DATA_SET
        self._data = data_override

    def get_data(self, deep_copy: bool = False):
        return copy.deepcopy(self._data) if deep_copy else self._data

    def read_file(self):
        raise NotImplemented

    def write_file(self):
        raise NotImplemented

    @classmethod
    def create(cls, location, read_now=True, data_override=None):
        from brain_brew.file_manager import FileManager
        _file_manager = FileManager.get_instance()
        formatted_location = cls.formatted_file_location(location)
        file = _file_manager.file_if_exists(formatted_location)

        if file is not None:
            return file

        file = cls(location, read_now=read_now, data_override=data_override)
        _file_manager.register_file(formatted_location, file)
        return file

    @classmethod
    def formatted_file_location(cls, location):
        return location

    @staticmethod
    def _sort_data(data, sort_by_keys, reverse_sort, case_insensitive_sort=None):
        if case_insensitive_sort is None:
            case_insensitive_sort = GlobalConfig.get_instance().flags.sort_case_insensitive

        if sort_by_keys:
            if case_insensitive_sort:
                def sort_method(i): return tuple((i[column] == "", i[column].lower()) for column in sort_by_keys)
            else:
                def sort_method(i): return tuple((i[column] == "", i[column]) for column in sort_by_keys)

            return sorted(data, key=sort_method, reverse=reverse_sort)
        elif reverse_sort:
            return list(reversed(data))

        return data



from collections import OrderedDict
from enum import Enum

from brain_brew.build_tasks.build_task_generic import BuildTaskGeneric
from brain_brew.constants.build_config_keys import BuildTaskEnum, BuildConfigKeys
from brain_brew.constants.crowdanki_keys import CAKeys
from brain_brew.constants.deckpart_keys import DeckPartNoteKeys
from brain_brew.file_manager import FileManager
from brain_brew.representation.generic.media_file import MediaFile
from brain_brew.utils import blank_str_if_none
from brain_brew.representation.generic.yaml_file import ConfigKey, YamlFile
from brain_brew.representation.json.crowd_anki_export import CrowdAnkiExport
from brain_brew.representation.json.deck_part_header import DeckPartHeader
from brain_brew.representation.json.deck_part_notemodel import CANoteModelKeys, DeckPartNoteModel
from brain_brew.representation.json.deck_part_notes import CANoteKeys, DeckPartNotes


class CrowdAnkiKeys(Enum):
    FILE = "file"
    NOTE_SORT_ORDER = "note_sort_order"
    MEDIA = "media"
    USELESS_NOTE_KEYS = "useless_note_keys"


class SourceCrowdAnki(YamlFile, BuildTaskGeneric):
    @staticmethod
    def get_build_keys():
        return [
            BuildTaskEnum("deck_parts_to_crowdanki", SourceCrowdAnki, "deck_parts_to_source", "source_to_deck_parts"),
            BuildTaskEnum("crowdanki_to_deck_parts", SourceCrowdAnki, "source_to_deck_parts", "deck_parts_to_source"),
        ]

    config_entry = {}
    expected_keys = {
        BuildConfigKeys.NOTES.value: ConfigKey(True, str, None),
        BuildConfigKeys.HEADERS.value: ConfigKey(True, str, None),

        CrowdAnkiKeys.FILE.value: ConfigKey(True, str, None),
        CrowdAnkiKeys.NOTE_SORT_ORDER.value: ConfigKey(False, list, None),
        CrowdAnkiKeys.MEDIA.value: ConfigKey(True, bool, None),
        CrowdAnkiKeys.USELESS_NOTE_KEYS.value: ConfigKey(True, dict, None)
    }
    subconfig_filter = None
    file_manager: FileManager

    headers: DeckPartHeader
    notes: DeckPartNotes

    crowd_anki_export: CrowdAnkiExport
    should_handle_media: bool
    useless_note_keys: dict

    def __init__(self, config_data: dict, read_now=True):
        self.setup_config_with_subconfig_replacement(config_data)
        self.verify_config_entry()

        self.file_manager = FileManager.get_instance()

        self.headers = DeckPartHeader.create(self.config_entry[BuildConfigKeys.HEADERS.value], read_now=read_now)
        self.notes = DeckPartNotes.create(self.config_entry[BuildConfigKeys.NOTES.value], read_now=read_now)
        self.crowd_anki_export = CrowdAnkiExport.create(self.config_entry[CrowdAnkiKeys.FILE.value], read_now=read_now)

        self.should_handle_media = self.config_entry[CrowdAnkiKeys.MEDIA.value]
        self.useless_note_keys = self.config_entry[CrowdAnkiKeys.USELESS_NOTE_KEYS.value]

    @classmethod
    def from_yaml(cls, yaml_file_name, read_now=True):
        config_data = YamlFile.read_file(yaml_file_name)

        return SourceCrowdAnki(config_data, read_now=read_now)

    def notes_to_deck_parts(self, notes_json, note_models_id_name_dict):
        for note in notes_json:
            for key in self.useless_note_keys:
                if key in note:
                    del note[key]
                # TODO: else raise error?

            if note[CANoteKeys.NOTE_MODEL.value] in note_models_id_name_dict:
                note[DeckPartNoteKeys.NOTE_MODEL.value] = note_models_id_name_dict[note[CANoteKeys.NOTE_MODEL.value]]
                del note[CANoteKeys.NOTE_MODEL.value]
            else:
                raise KeyError(f"Unknown NoteModel '{note[CANoteKeys.NOTE_MODEL.value]}'")

        return notes_json

    def source_to_deck_parts(self):
        source_data = self.crowd_anki_export.get_data(deep_copy=True)

        # Headers
        header_keys_to_ignore = {CAKeys.NOTE_MODELS.value, CAKeys.NOTES.value, CAKeys.MEDIA_FILES.value}

        headers_data = {key: source_data[key] for key in source_data if key not in header_keys_to_ignore}
        self.headers.set_data(headers_data)

        # Note Models
        note_models = [
            DeckPartNoteModel.create(model[CANoteModelKeys.NAME.value], data_override=model)
            for model in source_data[CAKeys.NOTE_MODELS.value]
        ]

        note_models_id_name_dict = {model.id: model.name for model in note_models}

        # Media
        for filename, file in self.crowd_anki_export.known_media.items():
            dp_media_file = self.file_manager.media_file_if_exists(filename)
            if dp_media_file:
                dp_media_file.set_override(file.source_loc)
            else:
                self.file_manager.new_media_file(filename, file.source_loc)

        # Notes
        notes_json = source_data[CAKeys.NOTES.value]
        notes_data = self.notes_to_deck_parts(notes_json, note_models_id_name_dict)

        self.notes.set_data(notes_data)

    def notes_to_source(self, note_models_dict_id_name):
        res_notes = self.notes.get_data(deep_copy=True)[DeckPartNoteKeys.NOTES.value]

        for note in res_notes:
            for key in self.useless_note_keys:
                note[key] = blank_str_if_none(self.useless_note_keys[key])

            note[CANoteKeys.NOTE_MODEL.value] = note_models_dict_id_name[note[DeckPartNoteKeys.NOTE_MODEL.value]]
            del note[DeckPartNoteKeys.NOTE_MODEL.value]

        return [OrderedDict(sorted(note.items())) for note in res_notes]

    def deck_parts_to_source(self):
        ca_json = {}

        # Headers
        ca_json.update(self.headers.get_data())

        # Media
        media_files = self.notes.referenced_media_files
        ca_json.setdefault(CAKeys.MEDIA_FILES.value, list(sorted([file.filename for file in media_files])))

        for file in media_files:
            filename = file.filename
            if filename in self.crowd_anki_export.known_media:
                self.crowd_anki_export.known_media[filename].set_override(file.source_loc)
            else:
                self.crowd_anki_export.known_media.setdefault(
                    filename, MediaFile(self.crowd_anki_export.media_loc + filename,
                                        filename, MediaFile.ManagementType.TO_BE_CLONED, file.source_loc)
                )

        # Note Models
        note_models = [DeckPartNoteModel.create(name) for name in self.notes.get_all_known_note_model_names()]

        ca_json.setdefault(CAKeys.NOTE_MODELS.value, [model.get_data() for model in note_models])

        note_models_dict_id_name = {model.name: model.id for model in note_models}

        # Notes
        ca_json.setdefault(CAKeys.NOTES.value, self.notes_to_source(note_models_dict_id_name))

        ordered_keys = OrderedDict(sorted(ca_json.items()))

        self.crowd_anki_export.set_data(ordered_keys)
